import typer
import os
import random
import warnings

from typing import List
from typing_extensions import Annotated

from ..annotation import Annotation
from ..genome import Genome

def split_callback(value:str):
    if value:
        return [item.strip() for item in value.split(",")]
    return []

app = typer.Typer(add_completion=False)
@app.command()
def main(
    annotation_file: Annotated[str, typer.Argument(
        help="Path to the input annotation GFF/GTF file."
    )],
    genome_file: Annotated[str, typer.Argument(
        help="Path to the input genome FASTA file."
    )] = "",
    chr_cap: Annotated[int, typer.Option(
        "--chr-cap", help="Add a chromosome cap to generate an annotation gff (and assembly fasta) subset(s)."
    )] = 2,
    chosen_chromosomes: Annotated[str, typer.Option(
        "-c", "--chromosomes", help="Overrides --chr-cap. Only the chosen chromosomes/scaffolds will be in the resulting annotation gff (and assembly fasta) subset(s). Add them and separate them by commas e.g. --chromosomes 'chr1,chr3'.",
    callback=split_callback
    )] = "",
    gene_cap: Annotated[int, typer.Option(
        "--gene-cap", help="Add a total gene number cap to reduce size of gff subset. The gene cap will affect scaffolds/chromosomes as uniformly as possible."
    )] = 3000,
    min_genes: Annotated[int, typer.Option(
        "--min-genes", help="Minimum total number of genes in the subset. Overrides --chr-cap if needed. Does not override --chosen-chromosomes if used."
    )] = 1500,
    annotation_name: Annotated[str, typer.Option(
        "-a", "--annotation-name", help="Annotation version, name or tag."
    )] = "{annotation-file}",
    genome_name: Annotated[str, typer.Option(
        "-g", "--genome-name", help="Genome assembly version, name or tag."
    )] = "{genome-file}",
    output_dir: Annotated[str, typer.Option(
        "-d", "--output-dir", help="Path to the output folder."
    )] = "./aegis_output/subsets/",
    output_annot_file: Annotated[str, typer.Option(
        "-oa", "--output-annot-file", help="Path to the output annotation filename, including extension."
    )] = "{annotation-name}_subset.gff3",
    output_genome_file: Annotated[str, typer.Option(
        "-og", "--output-genome-file", help="Path to the output genome filename, including extension."
    )] = "{genome-name}_subset.fasta",
    quiet: Annotated[bool, typer.Option(
        "-q", "--quiet", help="Keeps terminal reporting to a minimum."
    )] = False,
):
    """
    Obtain subsets of an annotation file, random or directed. Ramdom subsets prioritise chromosomal features if available. A lite version of a gff file and its corresponding genome fasta file can be useful for debugging/trialing tools.
    """
    chosen_chromosomes_set = set(chosen_chromosomes)

    if annotation_name == "{annotation-file}":
        annotation_name = os.path.splitext(os.path.basename(annotation_file))[0]

    if genome_name == "{genome-file}" and genome_file != "":
        genome_name = os.path.splitext(os.path.basename(genome_file))[0]
    elif genome_file == "":
        genome_name = "genome"

    os.makedirs(output_dir, exist_ok=True)

    if output_annot_file == "{annotation-name}_subset.gff3":
        output_annot_file = f"{annotation_name}_subset.gff3"

    if output_genome_file == "{genome-name}_subset.fasta":
        output_genome_file = f"{genome_name}_subset.fasta"

    if genome_file:
        g = Genome(genome_name, genome_file, quiet=quiet)
        a = Annotation(annotation_file, annotation_name, genome=g, quiet=quiet)
        common_chromosomes = set(a.chrs).intersection(set(g.scaffolds))
        common_actual_chromosomes = common_chromosomes - g.scaffold_names
        common_actual_chromosomes_minus_mt_chl = common_actual_chromosomes - g.accessory_chromosome_names

        if not common_chromosomes:
            raise ValueError(f"There are no common scaffolds/chromosomes between provided annotation and genome file.")
    else:
        a = Annotation(annotation_file, annotation_name, quiet=quiet)
        common_chromosomes = set(a.chrs)

    if not chosen_chromosomes_set:
        if chr_cap > len(common_chromosomes):
            chosen_chromosomes_set = common_chromosomes.copy()
            if genome_file:
                warnings.warn(f"Cap value {chr_cap} exceeds the number of available scaffolds/chrosomomes ({len(chosen_chromosomes_set)}) common to both genome and annotation files. The subset, in any case, will be based on common chromosomes/scaffolds.", category=UserWarning)
            else:
                warnings.warn(f"Cap value {chr_cap} exceeds the number of available scaffolds/chrosomomes ({len(chosen_chromosomes_set)}) in annotation file. The subset, in any case, will be based on common chromosomes/scaffolds.", category=UserWarning)
        elif chr_cap <= len(common_actual_chromosomes_minus_mt_chl):
            chosen_chromosomes_set = set(random.sample(list(common_actual_chromosomes_minus_mt_chl), chr_cap))
        elif chr_cap <= len(common_actual_chromosomes):
            chosen_chromosomes_set = set(random.sample(list(common_actual_chromosomes), chr_cap))
        else:
            chosen_chromosomes_set = set(random.sample(list(common_chromosomes), chr_cap))

        chosen_chromosomes_set = a.subset(chosen_features=chosen_chromosomes_set, gene_cap=gene_cap, common_chromosomes=common_chromosomes, min_genes=min_genes, quiet=quiet)
    
    else:
        # if chosen_chromosomes is selected min_genes parameter is ignored
        chosen_chromosomes_set = a.subset(chosen_features=chosen_chromosomes_set, gene_cap=gene_cap, common_chromosomes=common_chromosomes, min_genes=0, quiet=quiet)

    a.export.gff(custom_path=output_dir, tag=output_annot_file, subfolder=False, skip_atypical_fts=True)

    if genome_file:
        g.subset(chosen_features=chosen_chromosomes_set, quiet=quiet)
        g.export(output_folder=output_dir, file=output_genome_file, quiet=quiet)

if __name__ == "__main__":
    app()



"""
Tests for aegis.genefunctions — pure utility functions.
"""

import pytest

from pathlib import Path
from aegis.utils.genefunctions import (
    reverse_complement,
    find_ORFs,
    longest_ORF,
    trim_surplus,
    translate
)

from aegis.feature import Feature
from aegis.utils.gtf_gff import parse_gff_line, parse_gff_attributes
from aegis.utils.misc import count_occurrences, find_all_occurrences


# ============================================================
# parse_gff_attributes
# ============================================================

class TestParseGffAttributes:
    def test_standard_attributes(self):
        attrs = parse_gff_attributes("ID=gene1;Name=MyGene;Note=some note;Symbol=1,2;Alias=A1,A45")
        assert attrs["ID"] == "gene1"
        assert attrs["Name"] == ["MyGene"]
        assert attrs["Alias"] == ["A1", "A45"]
        assert attrs["Symbol"] == ["1", "2"]
        assert attrs["Note"] == "some note"

    def test_parent_attribute_splits_into_list(self):
        attrs = parse_gff_attributes("ID=exon1;Name=E123,E2;Parent=mRNA1,mRNA2")
        assert attrs["Name"] == ["E123", "E2"]
        assert attrs["Parent"] == ["mRNA1", "mRNA2"]

    def test_empty_string(self):
        assert parse_gff_attributes("") == {}

    def test_dot_string(self):
        assert parse_gff_attributes(".") == {}

    def test_single_attribute(self):
        attrs = parse_gff_attributes("ID=feat1")
        assert attrs["ID"] == "feat1"

    def test_derives_from_treated_as_parent(self):
        attrs = parse_gff_attributes("ID=cds1;Derives_from=mRNA1")
        assert attrs["Parent"] == ["mRNA1"]


# ============================================================
# parse_gff_line
# ============================================================

class TestParseGffLine:
    LINES_FILE = Path(__file__).resolve().parent / "test_data" / "input" / "annotation" / "genefunctions_lines.gff3"

    @classmethod
    def _read_line(cls, index):
        """Read a specific line (0-indexed) from genefunctions_lines.gff3."""
        with open(cls.LINES_FILE) as f:
            lines = f.read().splitlines()
        return lines[index]

    def test_basic_parsing(self):
        entry = parse_gff_line(self._read_line(0))  # gene line
        assert entry["ch"] == "chr1"
        assert entry["source"] == "aegis"
        assert entry["feature"] == "gene"
        assert entry["start"] == 1000
        assert entry["end"] == 5000
        assert entry["strand"] == "+"
        assert entry["id"] == "gene1"
        assert entry["pseudogene"] is False
        assert entry["transposable"] is False
        assert entry["decreasing_coordinates"] is False

    def test_pseudogene_detected(self):
        entry = parse_gff_line(self._read_line(1))  # pseudogene line
        assert entry["pseudogene"] is True

    def test_transposable_by_feature(self):
        entry = parse_gff_line(self._read_line(2))  # transposable_element_gene line
        assert entry["transposable"] is True

    def test_transposable_by_attribute(self):
        entry = parse_gff_line(self._read_line(3))  # gene with transposable=True attribute
        assert entry["transposable"] is True

    def test_decreasing_coordinates_swapped(self):
        entry = parse_gff_line(self._read_line(4))  # gene with start > end
        assert entry["decreasing_coordinates"] is True
        assert entry["start"] == 1000
        assert entry["end"] == 5000

    def test_multi_parent(self):
        entry = parse_gff_line(self._read_line(5))  # exon with two parents
        assert entry["parents"] == ["mRNA1", "mRNA2"]



# ============================================================
# count_occurrences
# ============================================================

class TestCountOccurrences:
    def test_basic_count(self):
        assert count_occurrences("AATTGGCC", "A") == 2
        assert count_occurrences("AATTGGCC", "G") == 2

    def test_missing_character(self):
        assert count_occurrences("AATTGGCC", "X") == 0

    def test_empty_string(self):
        assert count_occurrences("", "A") == 0


# ============================================================
# find_all_occurrences
# ============================================================

class TestFindAllOccurrences:
    def test_multiple_matches(self):
        matches = find_all_occurrences("AT", "ATATGCAT")
        # AT at pos 0, 2, 6
        assert len(matches) >= 2
        assert all(m[2] == "AT" for m in matches)

    def test_no_matches(self):
        matches = find_all_occurrences("ZZZ", "ATGCATGC")
        assert matches == []


# ============================================================
# reverse_complement
# ============================================================

class TestReverseComplement:
    def test_simple_sequence(self):
        # ATGC -> reverse complement = GCAT
        assert reverse_complement("ATGC") == "GCAT"

    def test_palindrome(self):
        # ATAT -> reverse complement = ATAT
        assert reverse_complement("ATAT") == "ATAT"

    def test_longer_sequence(self):
        seq = "ATGCGATCGA"
        rc = reverse_complement(seq)
        assert reverse_complement(rc) == seq  # double RC == original


# ============================================================
# find_ORFs / longest_ORF
# ============================================================

class TestFindORFs:
    def test_simple_orf(self):
        # ATG + 6 codons + TAA  = 24 nt
        seq = "ATGAAACCCGGGTTTTTTAACTAA"
        orfs = find_ORFs(seq)
        assert len(orfs) >= 1
        # the ORF should start with ATG and end with a stop codon
        orf_seq = orfs[0][0]
        assert orf_seq.startswith("ATG")
        assert orf_seq[-3:] in ("TAA", "TAG", "TGA")

    def test_no_start_codon(self):
        seq = "AAACCCGGGTTTTTTTAA"
        orfs = find_ORFs(seq)
        assert orfs == []

    def test_no_stop_codon_with_must_have_stop(self):
        seq = "ATGAAACCCGGG"  # no stop codon
        orfs = find_ORFs(seq, must_have_stop=True)
        assert orfs == []

    def test_no_stop_codon_without_must_have_stop(self):
        seq = "ATGAAACCCGGG"
        orfs = find_ORFs(seq, must_have_stop=False)
        assert len(orfs) >= 1


class TestLongestORF:
    def test_single_orf(self):
        orfs = [("ATGAAATAA", 0, 8)]
        result = longest_ORF(orfs)
        assert result[0] == "ATGAAATAA"

    def test_multiple_orfs(self):
        orfs = [
            ("ATGTAA", 0, 5),
            ("ATGAAAAAATAA", 10, 21),
            ("ATGCCCTAA", 30, 38),
        ]
        result = longest_ORF(orfs)
        assert result[0] == "ATGAAAAAATAA"

    def test_empty_list(self):
        result = longest_ORF([])
        assert result[0] == ""


# ============================================================
# trim_surplus
# ============================================================

class TestTrimSurplus:
    def test_divisible_by_3(self):
        seq = "ATGAAATAA"  # 9 nt, already divisible
        out, surplus = trim_surplus(seq)
        assert surplus is False
        assert out == seq

    def test_surplus_trimmed(self):
        seq = "ATGAAATAAG"  # 10 nt, surplus = 1
        out, surplus = trim_surplus(seq)
        assert surplus is True
        assert len(out) % 3 == 0


# ============================================================
# translate
# ============================================================

class TestTranslate:
    def test_translate_both_readthrough(self):
        # ATG AAA TAA -> M K *
        seq = "ATGAAATAA"
        start, end_stop, early_stop, surplus, gaps, protein, cs, ce = translate(seq, readthrough="both")
        assert start == "present"
        assert end_stop is True
        assert early_stop is False
        assert "M" in protein
        assert "*" in protein

    def test_translate_no_start(self):
        # GGG AAA TAA -> no M
        seq = "GGGAAATAA"
        start, end_stop, early_stop, surplus, gaps, protein, cs, ce = translate(seq, readthrough="both")
        assert start in ("absent", "late")

    def test_translate_none_readthrough_with_orfs(self):
        # has an ORF inside
        seq = "GGGGATGAAATAAGGG"
        # "none" mode finds the longest ORF
        result = translate(seq, readthrough="none")
        # returns a tuple
        assert isinstance(result, tuple)

    def test_translate_ambiguous_codons(self):
        # N in sequence signals ambiguity
        seq = "ATGNNATAA"
        start, end_stop, early_stop, surplus, gaps, protein, cs, ce = translate(seq, readthrough="both")
        assert gaps is True


# ============================================================
# overlap
# ============================================================

class TestOverlap:
    """Test the overlap function using simple mock objects."""

    class MockFeature(Feature):
        def __init__(self, start, end):
            super().__init__(feature_id="mock", ch="mock", source="mock", feature="mock", strand="mock", start=start, end=end,
                             score=".", phase=".", parents=[], attributes={})

    def test_overlapping_features(self):
        f1 = self.MockFeature(100, 300)
        f2 = self.MockFeature(200, 400)
        is_overlapping, bp = f1.overlap(f2)
        assert is_overlapping is True
        assert bp == 101  # 200..300 inclusive

    def test_non_overlapping_features(self):
        f1 = self.MockFeature(100, 200)
        f2 = self.MockFeature(300, 400)
        is_overlapping, bp = f1.overlap(f2)
        assert is_overlapping is False

    def test_adjacent_features(self):
        f1 = self.MockFeature(100, 200)
        f2 = self.MockFeature(201, 300)
        is_overlapping, bp = f1.overlap(f2)
        assert is_overlapping is False

    def test_contained_feature(self):
        f1 = self.MockFeature(100, 500)
        f2 = self.MockFeature(200, 300)
        is_overlapping, bp = f1.overlap(f2)
        assert is_overlapping is True

    def test_identical_features(self):
        f1 = self.MockFeature(100, 200)
        f2 = self.MockFeature(100, 200)
        is_overlapping, bp = f1.overlap(f2)
        assert is_overlapping is True
        assert bp == 101



import sys
import re

from ..conf import default_features, default_subfeatures
from collections import OrderedDict

def parse_gff_line(line):
    parts = line.strip().split("\t")

    # Interning high-frequency strings

    source = sys.intern(parts[1])
    feature = sys.intern(parts[2])
    strand = sys.intern(parts[6])
    phase = sys.intern(parts[7])

    if feature == "nucleotide_to_protein_match":
        ch = sys.intern(parts[0].split(":")[0])
    
    else:
        ch = sys.intern(parts[0])

    if "pseudo" in feature:
        pseudogene = True
    else:
        pseudogene = False

    start = int(parts[3])
    end = int(parts[4])

    if start > end:
        decreasing_coordinates = True
        actual_start, actual_end = (int(parts[4]), int(parts[3]))
    else:
        decreasing_coordinates = False
        actual_start, actual_end = (start, end)

    if "transposable" in feature or "transposon" in feature:
        transposable = True
    else:
        transposable = False

    transcript = False
    gene = False
    if feature in default_features["gene"]:
        gene = True        
    elif feature in default_features["transcript"]:
        transcript = True

    attr_dict = parse_gff_attributes(parts[8], gene, transcript)
    
    if not transposable:
        if attr_dict.get("transposable") == "True" or attr_dict.get("transposon") == "True":
            transposable = True

    entry = {
        "ch": ch,
        "source": source,
        "feature": feature,
        "start": actual_start,
        "end": actual_end,
        "score": parts[5],
        "strand": strand,
        "phase": phase,
        "attributes": attr_dict,
        "id": attr_dict.get("ID", ""),
        "parents": attr_dict.get("Parent", []),
        "pseudogene": pseudogene,
        "transposable": transposable,
        "decreasing_coordinates": decreasing_coordinates
    }

    if "ID" in attr_dict:
        del attr_dict["ID"]
    if "Parent" in attr_dict:
        del attr_dict["Parent"]

    entry["attributes"] = attr_dict

    if not entry["id"] and entry["parents"]:
        entry["id"] = f"{entry['parents'][0]}_{feature}_{actual_start}_{actual_end}"

    return entry

def parse_gff_attributes(attributes, gene:bool=False, transcript:bool=False):

    if not attributes or attributes == ".":
        return {}

    parsed = {}
    
    for p in attributes.split(";"):
        p = p.strip()
        if not p: 
            continue

        if "=" in p:
            key, val = p.split("=", 1)

            key = key.strip()
            val = val.strip()

            if key.lower() == "id" or gene and key.lower() in {"gene_id", "geneid"} or transcript and key.lower() in {"transcript_id", "transcriptid"}:
                mod_key = sys.intern("ID")
                parsed[mod_key] = val
            elif key.lower() in {"parent", "parents", "derives_from"} or key.lower() in {"gene_id", "geneid", "transcript_id", "transcriptid"}:
                mod_key = sys.intern("Parent")
                parsed[mod_key] = [x.strip() for x in val.split(",") if x.strip()]
            elif key.lower() == "alias":
                mod_key = sys.intern("Alias")
                parsed[mod_key] = [x.strip() for x in val.split(",") if x.strip()]
            elif key.lower() == "name":
                mod_key = sys.intern("Name")
                parsed[mod_key] = [x.strip() for x in val.split(",") if x.strip()]
            elif key.lower() == "symbol":
                mod_key = sys.intern("Symbol")
                parsed[mod_key] = [x.strip() for x in val.split(",") if x.strip()]
            else:
                key = sys.intern(key)
                parsed[key] = val

    return parsed

def parse_gtf_attributes(attr_string):
    """
    Parses the GFF/GTF attribute column and returns a dictionary.
    Handles GTF-specific format where values are quoted.
    """
    attributes = {}

    for match in re.finditer(r'(\w+)\s+"([^"]+)"', attr_string):
        key, value = match.groups()
        attributes[key] = value
    return attributes

def format_gff3_attributes(attrs, feature_type):
    """
    Formats a dictionary of attributes into a GFF3-compliant string.
    """

    special_keys = {'gene_id', 'transcript_id', 'exon_id', 'exon_number'}
    
    gff3_attrs = []

    if feature_type in default_features["gene"]:
        if 'gene_id' in attrs:
            gff3_attrs.append(f"ID={attrs['gene_id']}")
    
    elif feature_type in default_features["transcript"]:
        if 'transcript_id' in attrs:
            gff3_attrs.append(f"ID={attrs['transcript_id']}")
        if 'gene_id' in attrs:
            gff3_attrs.append(f"Parent={attrs['gene_id']}")
            
    elif feature_type in default_subfeatures:
        parent_id = f"{attrs['transcript_id']}"
        gff3_attrs.append(f"Parent={parent_id}")

        feature_id = f"{attrs['transcript_id']}"

        if 'exon_number' in attrs:
            feature_id += f"_e{attrs['exon_number']}"
            gff3_attrs.append(f"ID={feature_id}")
        elif feature_type in default_features["CDS"]:
            gff3_attrs.append(f"ID={feature_id}_CDS1")

    if 'gene_name' in attrs:
        gff3_attrs.append(f"Symbol={attrs['gene_name']}")
    elif 'transcript_name' in attrs:
        gff3_attrs.append(f"Symbol={attrs['transcript_name']}")

    for key, value in attrs.items():
        if key not in special_keys and key not in ['gene_name', 'transcript_name']:
            gff3_attrs.append(f"{key}={value}")
            
    return ";".join(gff3_attrs)

def convert_gtf_to_gff3(gtf_file, gff3_file, encoding, quiet:bool=False):
    """
    Reads a GTF file, converts it to GFF3 format, and writes to an output file.

    Uses a two-pass approach:
      Pass 1 — Collect all lines and infer gene/transcript boundaries from
               subfeature attributes (gene_id, transcript_id) when no explicit
               gene or transcript rows exist.
      Pass 2 — Emit gene, transcript, and subfeature lines in GFF3 format.
    """

    comments = []
    seen_genes = set()
    seen_transcripts = set()
    subfeature_lines = []

    inferred_genes: dict[str, dict] = OrderedDict()
    inferred_transcripts: dict[str, dict] = OrderedDict()

    explicit_gene_lines = []
    explicit_transcript_lines = []

    # PASS 1 — read file
    with open(gtf_file, 'r', encoding=encoding) as infile:
        for line in infile:
            if line.startswith('#'):
                if not line.startswith('##'):
                    comments.append(line)
                continue

            parts = line.strip().split('\t')
            if len(parts) != 9:
                if not quiet:
                    sys.stderr.write(f"Warning: Skipping malformed line: {line.strip()}\n")
                continue

            seqname, source, feature, start, end, score, strand, frame, attr_string = parts
            attributes = parse_gtf_attributes(attr_string)
            start_int = int(start)
            end_int = int(end)

            gene_id = attributes.get('gene_id', '')
            transcript_id = attributes.get('transcript_id', '')

            if feature in default_features["gene"]:
                if gene_id and gene_id not in seen_genes:
                    explicit_gene_lines.append((parts, attributes))
                    seen_genes.add(gene_id)
                continue

            if feature in default_features["transcript"]:
                if transcript_id and transcript_id not in seen_transcripts:
                    explicit_transcript_lines.append((parts, attributes))
                    seen_transcripts.add(transcript_id)
                continue

            subfeature_lines.append((parts, attributes, feature))

            if gene_id:
                if gene_id not in inferred_genes:
                    inferred_genes[gene_id] = {
                        'seqname': seqname, 'source': source,
                        'strand': strand, 'start': start_int,
                        'end': end_int, 'gene_name': attributes.get('gene_name', ''),
                        'gene_biotype': attributes.get('gene_biotype', ''),
                        'transcript_ids': OrderedDict()
                    }
                else:
                    g = inferred_genes[gene_id]
                    if start_int < g['start']:
                        g['start'] = start_int
                    if end_int > g['end']:
                        g['end'] = end_int

            if transcript_id:
                if transcript_id not in inferred_transcripts:
                    inferred_transcripts[transcript_id] = {
                        'gene_id': gene_id, 'seqname': seqname,
                        'source': source, 'strand': strand,
                        'start': start_int, 'end': end_int,
                    }
                else:
                    t = inferred_transcripts[transcript_id]
                    if start_int < t['start']:
                        t['start'] = start_int
                    if end_int > t['end']:
                        t['end'] = end_int

                if gene_id and gene_id in inferred_genes:
                    inferred_genes[gene_id]['transcript_ids'][transcript_id] = True

    # PASS 2 — write GFF3
    with open(gff3_file, 'w', encoding=encoding) as outfile:
        outfile.write("##gff-version 3\n")

        for c in comments:
            outfile.write(f"#{c}")

        for parts, attributes in explicit_gene_lines:
            gene_attrs = {'gene_id': attributes['gene_id']}
            if 'gene_name' in attributes:
                gene_attrs['gene_name'] = attributes['gene_name']
            if 'gene_biotype' in attributes:
                gene_attrs['gene_biotype'] = attributes['gene_biotype']
            gene_attr_str = format_gff3_attributes(gene_attrs, 'gene')
            outfile.write("\t".join([parts[0], parts[1], 'gene', parts[3], parts[4],
                                     parts[5], parts[6], parts[7], gene_attr_str]) + '\n')

        for gene_id, g in inferred_genes.items():
            if gene_id in seen_genes:
                continue
            gene_attrs = {'gene_id': gene_id}
            if g['gene_name']:
                gene_attrs['gene_name'] = g['gene_name']
            if g['gene_biotype']:
                gene_attrs['gene_biotype'] = g['gene_biotype']
            gene_attr_str = format_gff3_attributes(gene_attrs, 'gene')
            outfile.write("\t".join([g['seqname'], g['source'], 'gene',
                                     str(g['start']), str(g['end']), '.',
                                     g['strand'], '.', gene_attr_str]) + '\n')
            seen_genes.add(gene_id)

        for parts, attributes in explicit_transcript_lines:
            tx_attrs = {k: v for k, v in attributes.items()
                        if 'transcript' in k or 'gene' in k}
            tx_feature_type = 'transcript'
            if 'transcript_biotype' in attributes:
                if 'RNA' in attributes['transcript_biotype']:
                    tx_feature_type = attributes['transcript_biotype']
            tx_attr_str = format_gff3_attributes(tx_attrs, tx_feature_type)
            outfile.write("\t".join([parts[0], parts[1], tx_feature_type,
                                     parts[3], parts[4], parts[5], parts[6],
                                     parts[7], tx_attr_str]) + '\n')

        for t_id, t in inferred_transcripts.items():
            if t_id in seen_transcripts:
                continue
            tx_attrs = {'transcript_id': t_id}
            if t['gene_id']:
                tx_attrs['gene_id'] = t['gene_id']
            tx_attr_str = format_gff3_attributes(tx_attrs, 'mRNA')
            outfile.write("\t".join([t['seqname'], t['source'], 'mRNA',
                                     str(t['start']), str(t['end']), '.',
                                     t['strand'], '.', tx_attr_str]) + '\n')
            seen_transcripts.add(t_id)

        subfeature_counters: dict[tuple[str, str], int] = {}

        for parts, attributes, feature in subfeature_lines:
            transcript_id = attributes.get('transcript_id', '')
            gene_name = attributes.get('gene_name', '')

            gff3_attrs = []

            if transcript_id:
                key = (transcript_id, feature)
                subfeature_counters[key] = subfeature_counters.get(key, 0) + 1
                count = subfeature_counters[key]
                feature_id = f"{transcript_id}_{feature}{count}"
                gff3_attrs.append(f"ID={feature_id}")
                gff3_attrs.append(f"Parent={transcript_id}")
            else:
                gff3_attr_string = format_gff3_attributes(attributes, feature)
                outfile.write("\t".join([parts[0], parts[1], feature,
                                         parts[3], parts[4], parts[5], parts[6],
                                         parts[7], gff3_attr_string]) + '\n')
                continue

            if gene_name:
                gff3_attrs.append(f"Symbol={gene_name}")

            gff3_attr_string = ";".join(gff3_attrs)
            outfile.write("\t".join([parts[0], parts[1], feature,
                                     parts[3], parts[4], parts[5], parts[6],
                                     parts[7], gff3_attr_string]) + '\n')

    if not quiet:
        print(f"Successfully converted {gtf_file} to a gff format")

def detect_file_format(file_path, encoding, lines_to_check=20):
    """
    Detects if a file is likely GTF or GFF3 format.
    """
    try:
        with open(file_path, 'r', encoding=encoding) as f:

            i = 0

            for line in f:

                line = line.strip()
                if not line:
                    continue

                if line.startswith("##gff-version 3"):
                    return 'gff3'

                if line.startswith('#'):
                    continue

                i += 1

                if i >= lines_to_check:
                    break

                parts = line.split('\t')
                if len(parts) == 9:
                    attributes = parts[8]

                    if re.search(r'\w+=', attributes):
                        return 'gff3'

                    if re.search(r'\w+\s+"[^"]+";', attributes):
                        return 'gtf'
            
            # unknown format returns gff3
            return 'gff3'
            
    except Exception as e:
        sys.stderr.write(f"Error reading file for format detection: {e}\n")
        sys.exit(1)



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..annotation import Annotation

import time
import sys
import networkx as nx
import pandas as pd

from pathlib import Path
from tqdm import tqdm

from ..hits import OverlapHit

class AnnotationOverlaps:
    """
    Component for handling overlaps methods for the Annotation class.
    Accessed via 'annotation_object.overlaps'.
    """
    def __init__(self, annotation:Annotation):
        self._annot = annotation
        self.self = []
        self.other = []
    
    def detect(self, other:Annotation|None=None, sort_processes:int=1, clear=True, quiet:bool=True):
        """
        Detecting gene overlaps within the same annotation object or between
        annotation objects, provided they refer to the same genome.
        """
        start_time = time.time()
        # Check if stdout or stderr are redirected to files
        stdout_redirected = not sys.stdout.isatty()
        stderr_redirected = not sys.stderr.isatty()

        # Disable tqdm if stdout or stderr are redirected
        if stdout_redirected or stderr_redirected or quiet:
            disable = True
        else:
            disable = False

        if not self._annot.sorted:
            self._annot.sort_genes(processes=sort_processes)
        
        if other != None:
            if not other.sorted:
                other.sort_genes(processes=sort_processes)

        if clear:
            self.clear()
            if other != None:
                other.overlaps.clear()

        if other != None:

            if self._annot.genome == other.genome:

                if self._annot.genome == None:
                    if not quiet:
                        print(f"Note: Make sure that both annotations that are being compared are associated to the same genome version. Otherwise the resulting coordinate overlaps will not be correct.")
                
                if other.name in self.other or self._annot.name in other.overlaps.other:
                    print(f"Overlaps between {self._annot.id} and {other.id} "
                           "annotations have already been detected, please "
                           "run 'self.overlaps.clear()' and/or other.overlaps.clear() if you want to "
                           "recalculate them")
                else:
                    start_time = time.time()
                    progress_bar = tqdm(total=len(self._annot.all_gene_ids.keys()), disable=disable,
                                bar_format=(
                    f'\033[1;91mWorking out overlaps between {self._annot.id} and {other.id} annotations:\033[0m '
                    '{percentage:3.0f}%|'
                    f'\033[1;91m{{bar}}\033[0m| '
                    '{n}/{total} [{elapsed}<{remaining}]'))

                    if other.name not in self.other:
                        self.other.append(other.name)
                    if self._annot.name not in other.overlaps.other:
                        other.overlaps.other.append(self._annot.name)
                    for chr, genes in self._annot.chrs.items():
                        for g1 in genes.values():
                            progress_bar.update(1)
                            found_overlap = False
                            if chr in other.chrs:
                                for g2 in other.chrs[chr].values():
                                    overlapping, overlap_bp = g1.overlap(g2)
                                    if overlapping:
                                        found_overlap = True
                                    elif found_overlap and g1.end < g2.start:
                                        break
                                    else:
                                        continue
                                    if g1.strand == g2.strand:
                                        gene_orientation = True
                                    else:
                                        gene_orientation = False

                                    gene_query_percent = (overlap_bp / g1.size) * 100                  
                                    gene_target_percent = (overlap_bp / g2.size) * 100

                                    target_exons = False
                                    query_exons = False
                                    best_exon_overlap = 0
                                    exon_orientation = False
                                    overlapping = False
                                    for t1 in g1.transcripts.values():
                                        if t1.exons != []:
                                            query_exons = True
                                        for t2 in g2.transcripts.values():
                                            if t2.exons != []:
                                                target_exons = True
                                            if t1.strand == t2.strand:
                                                exon_orientation = True
                                            overlap_exon_temp = 0
                                            for e1 in t1.exons:
                                                for e2 in t2.exons:
                                                    overlap_temp, overlap_bp = e1.overlap(e2)
                                                    if overlap_temp:
                                                        overlap_exon_temp += overlap_bp
                                                        overlapping = True
                                            if overlap_exon_temp > best_exon_overlap:
                                                best_exon_overlap = overlap_exon_temp
                                                exon_query_size = t1.size
                                                exon_target_size = t2.size
                                    
                                    if target_exons and query_exons:
                                        exons_in_both = True
                                        if gene_orientation != exon_orientation:
                                            print(f"Warning: {self._annot.id} query and {other.id} target have discrepancies in the orientation of gene and exons. Genes: {g1.id} and {g2.id}.")
                                        if overlapping:
                                            exon_query_percent = (best_exon_overlap / exon_query_size) * 100
                                            exon_target_percent = (best_exon_overlap / exon_target_size) * 100
                                        else:
                                            exon_query_percent = 0
                                            exon_target_percent = 0
                                    else:
                                        exons_in_both = False
                                        exon_orientation = None
                                        exon_query_percent = None
                                        exon_target_percent = None


                                    target_CDS = False
                                    query_CDS = False
                                    best_CDS_overlap = 0
                                    CDS_orientation = False
                                    overlapping = False
                                    for t1 in g1.transcripts.values():
                                        if t1.CDSs != {}:
                                            query_CDS = True
                                        for t2 in g2.transcripts.values():
                                            if t2.CDSs != {}:
                                                target_CDS = True
                                            if t1.strand == t2.strand:
                                                CDS_orientation = True
                                            for CDS1 in t1.CDSs.values():
                                                for CDS2 in t2.CDSs.values():
                                                    overlap_CDS_temp = 0
                                                    for c1 in CDS1.CDS_segments:
                                                        for c2 in CDS2.CDS_segments:
                                                            overlap_temp, overlap_bp = c1.overlap(c2)
                                                            if not overlap_temp:
                                                                continue
                                                            overlap_CDS_temp += overlap_bp
                                                            overlapping = True
                                                    if overlap_CDS_temp > best_CDS_overlap:
                                                        best_CDS_overlap = overlap_CDS_temp
                                                        CDS_query_size = CDS1.size
                                                        CDS_target_size = CDS2.size
                                                
                                    if target_CDS and query_CDS:
                                        CDSs_in_both = True
                                        if gene_orientation != CDS_orientation:
                                            print(f"Warning: {self._annot.id} query and {other.id} target have discrepancies in the orientation of gene and CDS. Genes: {g1.id} and {g2.id}.")
                                        if overlapping:
                                            CDS_query_percent = (best_CDS_overlap / CDS_query_size) * 100
                                            CDS_target_percent = (best_CDS_overlap / CDS_target_size) * 100
                                        else:
                                            CDS_query_percent = 0
                                            CDS_target_percent = 0
                                    else:
                                        CDSs_in_both = False
                                        CDS_orientation = None
                                        CDS_query_percent = None
                                        CDS_target_percent = None


                                    protein_query_percent = None
                                    protein_target_percent = None
                                    if CDS_query_percent != None and CDS_query_percent != 0:
                                        if CDS_orientation:
                                            target_protein = False
                                            query_protein = False
                                            best_protein_overlap = 0
                                            overlapping = False
                                            for t1 in g1.transcripts.values():
                                                if t1.CDSs != {}:
                                                    query_protein = True
                                                for t2 in g2.transcripts.values():
                                                    if t2.CDSs != {}:
                                                        target_protein = True
                                                    for CDS1 in t1.CDSs.values():
                                                        for CDS2 in t2.CDSs.values():
                                                            overlap_protein_temp = 0
                                                            for c1 in CDS1.CDS_segments:
                                                                for c2 in CDS2.CDS_segments:
                                                                    overlap_temp, overlap_bp = c1.overlap(c2)
                                                                    if not overlap_temp:
                                                                        continue
                                                                    if c1.frame != c2.frame:
                                                                        continue
                                                                    
                                                                    overlap_protein_temp += overlap_bp
                                                                    overlapping = True
                                                            if overlap_protein_temp > best_protein_overlap:
                                                                best_protein_overlap = overlap_protein_temp
                                                                protein_query_size = CDS1.size
                                                                protein_target_size = CDS2.size
                                                        
                                            if target_protein and query_protein:
                                                if overlapping:
                                                    protein_query_percent = (best_protein_overlap / protein_query_size) * 100
                                                    protein_target_percent = (best_protein_overlap / protein_target_size) * 100
                                                else:
                                                    protein_query_percent = 0
                                                    protein_target_percent = 0

                                    g1.overlaps["other"].append(OverlapHit(g2.id, 
                                                                    other.name,
                                                                    gene_orientation,
                                                                    gene_query_percent,
                                                                    gene_target_percent,
                                                                    exons_in_both,
                                                                    exon_query_percent,
                                                                    exon_target_percent,
                                                                    CDSs_in_both,
                                                                    CDS_query_percent,
                                                                    CDS_target_percent,
                                                                    protein_query_percent,
                                                                    protein_target_percent,
                                                                    g2.conserved_synteny,
                                                                    g2.extra_copy))

                                    g2.overlaps["other"].append(OverlapHit(g1.id,
                                                                    self._annot.name,
                                                                    gene_orientation,
                                                                    gene_target_percent,
                                                                    gene_query_percent,
                                                                    exons_in_both,
                                                                    exon_target_percent,
                                                                    exon_query_percent,
                                                                    CDSs_in_both,
                                                                    CDS_target_percent,
                                                                    CDS_query_percent,
                                                                    protein_query_percent,
                                                                    protein_target_percent,
                                                                    g1.conserved_synteny,
                                                                    g1.extra_copy))
                    self._annot.add_aliases()
                    other.add_aliases()
                    now = time.time()
                    lapse = now - start_time
                    progress_bar.close()
                    if not quiet:
                        print(f"\nDetecting overlaps between {other.id} and {self._annot.id} annotations took {round(lapse/60, 1)} minutes")
            else:
                print(f"Did not generate overlaps between {other.id} and {self._annot.id} annotations as they are associated to different genomes")

        else:
            if self.self != []:
                print("There are already detected 'self' gene overlaps, please run 'self.clear()' if you want to recalculate them")
            else:
                progress_bar = tqdm(total=len(self._annot.all_gene_ids.keys()), disable=disable,
                            bar_format=(
                f'\033[1;91mWorking out overlaps within {self._annot.id} annotation:\033[0m '
                '{percentage:3.0f}%|'
                f'\033[1;91m{{bar}}\033[0m| '
                '{n}/{total} [{elapsed}<{remaining}]'))

                # making sure self overlaps are not added twice
                start_time = time.time()
                self.self = set(self.self)
                for chr, genes in self._annot.chrs.items():
                    gl = list(genes.keys())[1:]
                    for g1 in genes.values():
                        progress_bar.update(1)
                        found_overlap = False
                        for gl_id in gl:
                            g2 = genes[gl_id]
                            if g1.id == g2.id:
                                continue
                            overlapping, overlap_bp = g1.overlap(g2)

                            if overlapping:
                                self.self.add(g1.id)
                                self.self.add(g2.id)
                                found_overlap = True

                            elif found_overlap and g1.end < g2.start:
                                break
                            else:
                                continue

                            if g1.strand == g2.strand:
                                gene_orientation = True
                            else:
                                gene_orientation = False

                            gene_query_percent = (overlap_bp / g1.size) * 100                  
                            gene_target_percent = (overlap_bp / g2.size) * 100

                            target_exons = False
                            query_exons = False
                            best_exon_overlap = 0
                            exon_orientation = False
                            overlapping = False
                            for t1 in g1.transcripts.values():
                                if t1.exons != []:
                                    query_exons = True
                                for t2 in g2.transcripts.values():
                                    if t2.exons != []:
                                        target_exons = True
                                    if t1.strand == t2.strand:
                                        exon_orientation = True
                                    overlap_exon_temp = 0
                                    for e1 in t1.exons:
                                        for e2 in t2.exons:
                                            overlap_temp, overlap_bp = e1.overlap(e2)
                                            if overlap_temp:
                                                overlap_exon_temp += overlap_bp
                                                overlapping = True
                                    if overlap_exon_temp > best_exon_overlap:
                                        best_exon_overlap = overlap_exon_temp
                                        exon_query_size = t1.size
                                        exon_target_size = t2.size
                            
                            if target_exons and query_exons:
                                exons_in_both = True
                                if gene_orientation != exon_orientation:
                                    print(f"Warning: {self._annot.id} query and target have discrepancies in the orientation of gene and exons. Genes: {g1.id} and {g2.id}")
                                if overlapping:
                                    exon_query_percent = (best_exon_overlap / exon_query_size) * 100
                                    exon_target_percent = (best_exon_overlap / exon_target_size) * 100
                                else:
                                    exon_query_percent = 0
                                    exon_target_percent = 0
                            else:
                                exons_in_both = False
                                exon_orientation = None
                                exon_query_percent = None
                                exon_target_percent = None

                            target_CDS = False
                            query_CDS = False
                            best_CDS_overlap = 0
                            CDS_orientation = False
                            overlapping = False
                            for t1 in g1.transcripts.values():
                                if t1.CDSs != {}:
                                    query_CDS = True
                                for t2 in g2.transcripts.values():
                                    if t2.CDSs != {}:
                                        target_CDS = True
                                    if t1.strand == t2.strand:
                                        CDS_orientation = True
                                    for CDS1 in t1.CDSs.values():
                                        for CDS2 in t2.CDSs.values():
                                            overlap_CDS_temp = 0
                                            for c1 in CDS1.CDS_segments:
                                                for c2 in CDS2.CDS_segments:
                                                    overlap_temp, overlap_bp = c1.overlap(c2)
                                                    if not overlap_temp:
                                                        continue
                                                    overlap_CDS_temp += overlap_bp
                                                    overlapping = True
                                            if overlap_CDS_temp > best_CDS_overlap:
                                                best_CDS_overlap = overlap_CDS_temp
                                                CDS_query_size = CDS1.size
                                                CDS_target_size = CDS2.size
                                        
                            if target_CDS and query_CDS:
                                CDSs_in_both = True
                                if gene_orientation != CDS_orientation:
                                    print(f"Error: {self._annot.id} query and target have discrepancies in the orientation of gene and CDS. Genes: {g1.id} and {g2.id}. DO NOT CONTINUE! -> fix the problem!")
                                if overlapping:
                                    CDS_query_percent = (best_CDS_overlap / CDS_query_size) * 100
                                    CDS_target_percent = (best_CDS_overlap / CDS_target_size) * 100
                                else:
                                    CDS_query_percent = 0
                                    CDS_target_percent = 0
                            else:
                                CDSs_in_both = False
                                CDS_orientation = None
                                CDS_query_percent = None
                                CDS_target_percent = None

                            if CDSs_in_both:
                                protein_query_percent = 0
                                protein_target_percent = 0
                                if CDS_query_percent != None and CDS_query_percent != 0:
                                    if CDS_orientation:
                                        target_protein = False
                                        query_protein = False
                                        best_protein_overlap = 0
                                        overlapping = False
                                        for t1 in g1.transcripts.values():
                                            if t1.CDSs != {}:
                                                query_protein = True
                                            for t2 in g2.transcripts.values():
                                                if t2.CDSs != {}:
                                                    target_protein = True
                                                for CDS1 in t1.CDSs.values():
                                                    for CDS2 in t2.CDSs.values():
                                                        overlap_protein_temp = 0
                                                        for c1 in CDS1.CDS_segments:
                                                            for c2 in CDS2.CDS_segments:
                                                                overlap_temp, overlap_bp = c1.overlap(c2)
                                                                if not overlap_temp:
                                                                    continue
                                                                if c1.frame != c2.frame:
                                                                    continue
                                                                
                                                                overlap_protein_temp += overlap_bp
                                                                overlapping = True
                                                        if overlap_protein_temp > best_protein_overlap:
                                                            best_protein_overlap = overlap_protein_temp
                                                            protein_query_size = CDS1.size
                                                            protein_target_size = CDS2.size
                                                    
                                        if target_protein and query_protein:
                                            if overlapping:
                                                protein_query_percent = (best_protein_overlap / protein_query_size) * 100
                                                protein_target_percent = (best_protein_overlap / protein_target_size) * 100
                            else:

                                protein_query_percent = None
                                protein_target_percent = None

                            g1.overlaps["self"].append(OverlapHit(g2.id, self._annot.name,
                                                                gene_orientation,
                                                                gene_query_percent,
                                                                gene_target_percent,
                                                                exons_in_both,
                                                                exon_query_percent,
                                                                exon_target_percent,
                                                                CDSs_in_both,
                                                                CDS_query_percent,
                                                                CDS_target_percent,
                                                                protein_query_percent,
                                                                protein_target_percent,
                                                                g2.conserved_synteny,
                                                                g2.extra_copy))

                            g2.overlaps["self"].append(OverlapHit(g1.id, self._annot.name,
                                                                gene_orientation,
                                                                gene_target_percent,
                                                                gene_query_percent,
                                                                exons_in_both,
                                                                exon_target_percent,
                                                                exon_query_percent,
                                                                CDSs_in_both,
                                                                CDS_target_percent,
                                                                CDS_query_percent,
                                                                protein_query_percent,
                                                                protein_target_percent,
                                                                g1.conserved_synteny,
                                                                g1.extra_copy))
                        try:
                            gl.remove(g1.id)
                        except:
                            pass

                self.self = list(self.self)
                progress_bar.close()
                now = time.time()
                lapse = now - start_time
                if not quiet:
                    print(f"\nDetecting gene overlaps within the {self._annot.id} annotation took {round(lapse/60, 1)} minutes\n")
                    print(f"\nThere are {len(self.self)} genes overlapping with other genes in {self._annot.id} annotation\n")
                self.add_qualitative_info()

    def as_networks(self, self_mode:bool=True):
        self.networks = {}
        for chr, genes in self._annot.chrs.items():
            G = nx.Graph()
            for g in genes.values():
                if self_mode:
                    overlaps = g.overlaps["self"]
                else:
                    overlaps = g.overlaps["other"]
                for o in overlaps:
                    G.add_edge(g.id, o.id)
            self.networks[chr] = list(nx.connected_components(G))


    def clear(self, keep_self=False, keep_other=False):
        if not keep_self:
            self.self = []
            for genes in self._annot.chrs.values():
                for g in genes.values():
                    g.overlaps["self"] = []
        if not keep_other:
            self.other = []
            for genes in self._annot.chrs.values():
                for g in genes.values():
                    g.overlaps["other"] = []

    def add_qualitative_info(self, quiet:bool=True):
        """
        Number of unique full segment overlaps between genes including all transcript variants.
        """
        # Check if stdout or stderr are redirected to files
        stdout_redirected = not sys.stdout.isatty()
        stderr_redirected = not sys.stderr.isatty()

        # Disable tqdm if stdout or stderr are redirected
        if stdout_redirected or stderr_redirected or quiet:
            disable = True
        else:
            disable = False
        total = len(self._annot.all_gene_ids.keys())

        progress_bar = tqdm(total=total, disable=disable,
                                bar_format=(
                    f'\033[1;95mAdding qualitative info to {self._annot.id} overlaps:\033[0m '
                    '{percentage:3.0f}%|'
                    f'\033[1;95m{{bar}}\033[0m| '
                    '{n}/{total} [{elapsed}<{remaining}]'))
        for genes in self._annot.chrs.values():
            for g in genes.values():
                progress_bar.update(1)
                for o in g.overlaps["self"]:
                    if o.exon_query_percent > 0:

                        g_exons, g_CDSs, g_UTRs = set(), set(), set()
                        o_exons, o_CDSs, o_UTRs = set(), set(), set()

                        for t in g.transcripts.values():
                            if t.main:
                                g_exons.update(set([(e.start, e.end) for e in t.exons]))
                                for c in t.CDSs.values():
                                    if c.main:
                                        g_CDSs.update(set([(cs.start, cs.end, cs.frame) for cs in c.CDS_segments]))
                                        g_UTRs.update(set([(u.start, u.end) for u in c.UTRs]))
                                
                        chrom = self._annot.all_gene_ids[o.id]
                        for t in self._annot.chrs[chrom][o.id].transcripts.values():
                            if t.main:
                                o_exons.update(set([(e.start, e.end) for e in t.exons]))
                                for c in t.CDSs.values():
                                    if c.main:
                                        o_CDSs.update(set([(cs.start, cs.end, cs.frame) for cs in c.CDS_segments]))
                                        o_UTRs.update(set([(u.start, u.end) for u in c.UTRs]))

                        for e1 in g_exons:
                            for e2 in o_exons:
                                if e1 == e2:
                                    o.full_exon_overlaps += 1

                        for c1 in g_CDSs:
                            for c2 in o_CDSs:
                                if c1[0] == c2[0] and c1[1] == c2[1]:
                                    o.full_CDS_overlaps += 1
                                    if c1[2] == c2[2]:
                                        o.full_protein_overlaps += 1

                        for u1 in g_UTRs:
                            for u2 in o_UTRs:
                                if u1 == u2:
                                    o.full_UTR_overlaps += 1

        progress_bar.close()

    def clear_with_selected_CDSs(self):
        for genes in self._annot.chrs.values():
            for g in genes.values():
                g.overlap_with_selected_CDS = False   

    def clear_with_selected_exons(self):
        for genes in self._annot.chrs.values():
            for g in genes.values():
                g.overlap_with_selected_exon = False

    def export(self, custom_path: str = "", overlap_threshold: int = 6, verbose: bool = True, synteny: bool = False, NAs: bool = True, export_csv: bool = False, export_self: bool = False, output_file: str = "", quiet: bool = False, copies_info: bool = False) -> pd.DataFrame:
        start_time = time.time()
        if export_self:
            export = "self"
            export_tag = "self_"
        else:
            export = "other"
            export_tag = ""

        tag = f"{export_tag}{self._annot.id}{self._annot.feature_suffix}_overlap_t{overlap_threshold}"
        

        if custom_path:
            export_folder = Path(custom_path)
        else:
            export_folder = Path(custom_path or self._annot.path) / "overlaps"
        export_folder.mkdir(parents=True, exist_ok=True)
        export_folder = str(export_folder) + "/"

        correct_order = ["gene_id_A", "gene_id_B", "gene_id_A_synteny_conserved", "gene_id_B_synteny_conserved", "same_strand", "min_gene_percent", "min_exon_percent", "min_CDS_percent", "gene_id_A_origin", "gene_id_B_origin", "overlap_score", "gene_id_A_copy", "gene_id_B_copy"]

        rows = []

        for genes in self._annot.chrs.values():
            for g in genes.values():
                for name, hits in g.overlaps.items():
                    if name == export:
                        for hit in hits:
                            if hit.score < overlap_threshold:
                                continue

                            row = {
                                "gene_id_A": g.id,
                                "gene_id_B": hit.id,
                                "gene_id_A_origin": self._annot.name,
                                "gene_id_B_origin": hit.origin,
                                "overlap_score": hit.score,
                            }

                            if synteny:
                                row["gene_id_A_synteny_conserved"] = g.conserved_synteny if self._annot.liftover else pd.NA
                                row["gene_id_B_synteny_conserved"] = hit.target_synteny_conserved

                            if verbose:
                                row["same_strand"] = hit.orientation
                                row["min_gene_percent"] = hit.min_gene_percent
                                row["min_exon_percent"] = hit.min_exon_percent
                                row["min_CDS_percent"] = hit.min_CDS_percent

                            if copies_info:
                                row["gene_id_A_copy"] = g.extra_copy
                                row["gene_id_B_copy"] = hit.extra_copy

                            rows.append(row)

        eq_df = pd.DataFrame(rows)


        if eq_df.empty:

            if export == "self":
                print(f"\nNo {self._annot.id} self overlaps were detected.")
            else:
                print(f"\nNo {self._annot.id} overlaps to the following annotation(s) '{self.other}' were detected.")
            cols = [
                "gene_id_A",
                "gene_id_B",
                "gene_id_A_origin",
                "gene_id_B_origin",
                "overlap_score",
            ]

            if synteny:
                cols += [
                    "gene_id_A_synteny_conserved",
                    "gene_id_B_synteny_conserved",
                ]

            if verbose:
                cols += [
                    "same_strand",
                    "min_gene_percent",
                    "min_exon_percent",
                    "min_CDS_percent",
                ]

            if copies_info:
                cols += [
                    "gene_id_A_copy",
                    "gene_id_B_copy",
                ]
            rows = []
            eq_df = pd.DataFrame(rows, columns=cols)



        if export == "self":
            eq_df["sorted_id_pair"] = eq_df.apply(lambda row: tuple(sorted([row["gene_id_A"], row["gene_id_B"]])), axis=1)
            eq_df = eq_df.drop_duplicates(subset="sorted_id_pair").drop(columns="sorted_id_pair")
            eq_df.drop(inplace=True, columns=["gene_id_A_origin", "gene_id_B_origin"])

        else:
            eq_df["sorted_id_pair"] = eq_df.apply(lambda row: tuple(sorted([f'{row["gene_id_A"]}_{row["gene_id_A_origin"]}', f'{row["gene_id_B"]}_{row["gene_id_B_origin"]}'])), axis=1)
            eq_df = eq_df.drop_duplicates(subset="sorted_id_pair").drop(columns="sorted_id_pair")

        if NAs:
            tag += "_gene_id_A_NAs"
            if export == "self":
                overlapping_genes = set(pd.concat([eq_df["gene_id_A"], eq_df["gene_id_B"]]).dropna())
            else:
                overlapping_genes = set(eq_df["gene_id_A"].dropna())

            na_rows = []

            if not copies_info:

                if export == "self":

                    for genes in self._annot.chrs.values():
                        for g in genes.values():
                            if g.id not in overlapping_genes:
                                na_rows.append({
                                    "gene_id_A": g.id,
                                    "overlap_score": 0,
                                    "gene_id_A_synteny_conserved": g.conserved_synteny if self._annot.liftover else pd.NA
                                })

                else:

                    for genes in self._annot.chrs.values():
                        for g in genes.values():
                            if g.id not in overlapping_genes:
                                na_rows.append({
                                    "gene_id_A": g.id,
                                    "gene_id_A_origin": self._annot.name,
                                    "overlap_score": 0,
                                    "gene_id_A_synteny_conserved": g.conserved_synteny if self._annot.liftover else pd.NA
                                })

                if synteny:
                    for g_id in self._annot.unmapped:
                        na_rows.append({
                            "gene_id_A": g_id,
                            "gene_id_A_origin": self._annot.name
                        })

            else:

                if export == "self":

                    for genes in self._annot.chrs.values():
                        for g in genes.values():
                            if g.id not in overlapping_genes:
                                na_rows.append({
                                    "gene_id_A": g.id,
                                    "gene_id_A_copy": g.extra_copy,
                                    "overlap_score": 0,
                                    "gene_id_A_synteny_conserved": g.conserved_synteny if self._annot.liftover else pd.NA
                                })

                else:

                    for genes in self._annot.chrs.values():
                        for g in genes.values():
                            if g.id not in overlapping_genes:
                                na_rows.append({
                                    "gene_id_A": g.id,
                                    "gene_id_A_origin": self._annot.name,
                                    "overlap_score": 0,
                                    "gene_id_A_copy": g.extra_copy,
                                    "gene_id_A_synteny_conserved": g.conserved_synteny if self._annot.liftover else pd.NA
                                })

                if synteny:
                    for g_id in self._annot.unmapped:
                        na_rows.append({
                            "gene_id_A": g_id,
                            "gene_id_A_origin": self._annot.name
                        })          

            # Combine with the original df
            if na_rows:
                eq_df = pd.concat([eq_df, pd.DataFrame(na_rows)], ignore_index=True)

        eq_df = eq_df[[col for col in correct_order if col in eq_df.columns]]

        if synteny:
            column_sort_order = ["gene_id_A_origin", "gene_id_B_origin", "overlap_score", "gene_id_A_synteny_conserved", "gene_id_B_synteny_conserved", "gene_id_A", "gene_id_B"]
            ascending = [True, True, False, False, False, True, True]
        elif export == "self":
            column_sort_order = ["overlap_score", "gene_id_A", "gene_id_B"]
            ascending = [False, True, True]
        else:
            column_sort_order = ["gene_id_A_origin", "gene_id_B_origin", "overlap_score", "gene_id_A", "gene_id_B"]
            ascending = [True, True, False, True, True]            

        eq_df.sort_values(by=column_sort_order, ascending=ascending, inplace=True)
        eq_df.reset_index(drop=True, inplace=True)
        
        if export_csv:
            if output_file:
                export_path = f"{export_folder}{output_file}"
            else:
                export_path = f"{export_folder}{tag}.csv"

            eq_df.to_csv(export_path, sep="\t", index=False, na_rep="NA")

            now = time.time()
            lapse = now - start_time
            if not quiet:
                if export == "self":
                    print(f"\nExporting {self._annot.id} self overlaps took {round(lapse/60, 1)} minutes")
                else:
                    print(f"\nExporting {self._annot.id} overlaps to the following annotation(s) '{self.other}' took {round(lapse/60, 1)} minutes")
        
        return eq_df


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Saturday March 23 14:29:53 2023

@authors: David Navarro, Antonio Santiago
"""

import colorlover as cl
import plotly.graph_objects as go

def hex_to_rgb(hex_string):
    hex_string = hex_string.lstrip("#")

    return tuple(int(hex_string[i:i + 2], 16) for i in (0, 2, 4))

palettes = {}
palettes["extreme"] = ["#8B0000", "#ADD8E6"]
palettes["purple"] = ["#F7B7A3", "#EA5F89", "#9B3192", "#57167E", "#2B0B3F"]
palettes["pastel"] = ["#A8D6B5", "#FFC9A5", "#C9CCE5", "#EFCEDF", "#DFF2B9"]

common_font = dict(family="Arial", size=18)

def pie_chart(labels:list[str], values:list[int], export_folder:str, tag:str, title:str, hovertext_labels:list|None=None, palette_name:str="purple"):
    colours = palettes[palette_name]
    fig = go.Figure(data=[go.Pie(labels=labels, sort=False, values=values, hoverinfo="label+value", textfont=common_font, hole=0.4,textinfo="percent")])
    fig.update_layout(title=title, title_font=dict(family="Arial", size=22), hoverlabel=dict(font_size=common_font["size"], font_family=common_font["family"]), title_x=0.5, legend=dict(font=common_font))
    if len(colours) == len(labels):
        pass
    elif len(colours) > len(labels):
        colours = colours[:len(labels)]
    else:
        rgb_colours = []
        for colour in colours:
            rgb_colours.append(hex_to_rgb(colour))
        colours = cl.interp(rgb_colours, len(labels))
    fig.update_traces(marker=dict(colors=colours, line=dict(color='#000000', width=0.05)))

    if hovertext_labels != None:
        fig.update_traces(hoverinfo='value+text', text=hovertext_labels)
        fig.update_layout(hoverlabel=dict(font_size=12))

    total = sum(values)
    fig.add_annotation(text=f'Total: {total}', x=0.5, y=0.5, font=common_font, showarrow=False)

    fig.write_image(f"{export_folder}{tag}.pdf")
    fig.write_html(f"{export_folder}{tag}.html")

def barplot(values:list[int], export_folder:str, tag:str, title:str, max_x:int|None=None):

    if max_x == None:
        max_x = max(values) + 1
    else:
        tag += f"_max_x_{max_x}"

    min_x = 0
    if len(values) > 0:
        if min(values) >= 0:
            min_x = 0
        else:
            min_x = min(values) - 1
    x_range = [min_x, max_x]

    fig = go.Figure(data=[go.Histogram(x=values, marker_color='#4CAF50', marker_line_color='#000000', marker_line_width=0.05)])
    fig.update_layout(title=title, title_x=0.5, xaxis_title=tag, yaxis_title="Frequency", xaxis=dict(range=x_range), bargap=0.05, bargroupgap=0.1)
    fig.update_layout(title_font=dict(family="Arial", size=22), hoverlabel=dict(font_size=common_font["size"], font_family=common_font["family"]))
    fig.write_image(f"{export_folder}{tag}.pdf")
    fig.write_html(f"{export_folder}{tag}.html")



import typer

from .extract import main as extract_main
from .merge import main as merge_main
from .motif_search import main as motifs_main
from .orthology import main as orthology_main
from .overlap import main as overlap_main
from .prune import main as prune_main
from .reformat import main as reformat_main
from .rename import main as rename_main
from .subset import main as subset_main
from .summary import main as summary_main
from .symbols import main as symbols_main
from .tidy import main as tidy_main
from .tidy_genome import main as tidy_genome_main
from . import list as list_cmd

app = typer.Typer(
    name="aegis",
    help="Annotation Extraction Genomic Integration Suite.",
    add_completion=False,
    no_args_is_help=True,
)

# Single-command modules: register their main function directly as a command
app.command(name="extract", help="Extract sequences from a genome based on an annotation file.")(extract_main)
app.command(name="merge", help="Merge two annotation files.")(merge_main)
app.command(name="motifs", help="Search for DNA motifs in promoter regions.")(motifs_main)
app.command(name="orthology", help="Pairwise orthology analysis between two annotations from different genome assembly.")(orthology_main)
app.command(name="overlap", help="Detect overlaps between annotations associated to the same genome assembly.")(overlap_main)
app.command(name="prune", help="Remove chosen gene or transcript models from an annotation.")(prune_main)
app.command(name="reformat", help="Reformat an annotation file between GFF and GTF formats.")(reformat_main)
app.command(name="rename", help="Rename gene/transcript/subfeature IDs in an annotation.")(rename_main)
app.command(name="subset", help="Subset an annotation, and optionally its corresponding genome assembly, in various different ways. May your testing be lite.")(subset_main)
app.command(name="summary", help="Output summary statistics for an annotation.")(summary_main)
app.command(name="symbols", help="Add gene symbols to an annotation.")(symbols_main)
app.command(name="tidy", help="Clean and reformat an annotation file, allowing customisable format flavours.")(tidy_main)
app.command(name="tidy-genome", help="Clean and reformat a genome FASTA file.")(tidy_genome_main)

# Multi-command module: keep as a sub-group
app.add_typer(list_cmd.app, name="list", help="List genes or transcripts from an annotation file.")

if __name__ == "__main__":
    app()



"""
Tests for aegis.misc_features — Protein and Promoter classes.
"""

import pytest

from aegis.misc_features import Protein, Promoter
from aegis.feature import Feature


# ============================================================
# Protein
# ============================================================

class TestProtein:
    def test_standard_protein(self):
        # ATG AAA TAA  ->  M K *
        nuc = "ATGAAATAA"
        p = Protein("prot1", nuc, "chr1")
        assert p.id == "prot1"
        assert p.ch == "chr1"
        assert "M" in p.seq
        assert p.start == "present"
        assert p.end_stop is True
        assert p.size > 0

    def test_partial_protein_no_start(self):
        # GGG AAA TAA  -> no M at start
        nuc = "GGGAAATAA"
        p = Protein("prot2", nuc, "chr1")
        assert p.partial is True
        assert p.start in ("absent", "late")

    def test_truncated_protein(self):
        # ATG TAA AAA TAA -> early stop at position 2
        nuc = "ATGTAAAAATAA"
        p = Protein("prot3", nuc, "chr1")
        assert p.truncated is True
        assert p.early_stop is True

    def test_protein_copy(self):
        nuc = "ATGAAATAA"
        p = Protein("prot1", nuc, "chr1")
        p2 = p.copy()
        p2.id = "changed"
        assert p.id == "prot1"

    def test_blast_hits_initially_empty(self):
        nuc = "ATGAAATAA"
        p = Protein("prot1", nuc, "chr1")
        assert p.blast_hits == []


# ============================================================
# Promoter
# ============================================================

class TestPromoter:
    def test_init(self):
        pr = Promoter(
            promoter_type="standard",
            feature_id="prom1",
            ch="chr1",
            source="aegis",
            feature="promoter",
            strand="+",
            start=2000,
            end=4000,
            score=".",
            phase="."
        )
        assert pr.type == "standard"
        assert pr.id == "prom1"
        assert pr.size == 2001
        assert isinstance(pr, Feature)

    def test_promoter_atg_type(self):
        pr = Promoter(
            promoter_type="ATG",
            feature_id="prom2",
            ch="chr1",
            source="aegis",
            feature="promoter",
            strand="-",
            start=10000,
            end=12000,
            score=".",
            phase="."
        )
        assert pr.type == "ATG"
        assert pr.strand == "-"



import typer
import os

from typing_extensions import Annotated

from ..annotation import Annotation

app = typer.Typer(add_completion=False, no_args_is_help=True)

@app.command()
def genes(
    annotation_file: Annotated[str, typer.Argument(
        help="Path to the input annotation GFF/GTF file."
    )],
    annotation_name: Annotated[str, typer.Option(
        "-a", "--annotation-name", help="Annotation version, name or tag. [default: derived from filename]"
    )] = "{annotation-file}",
    output_dir: Annotated[str, typer.Option(
        "-d", "--output-dir", help="Path to the output folder."
    )] = "./aegis_output/",
    output_file: Annotated[str, typer.Option(
        "-o", "--output-file", help="Path to the output filename, without extension."
    )] = "{annotation-name}_genes_list.tsv",
    lengths: Annotated[bool, typer.Option(
        "-l", "--lengths", help="Include feature lengths in the output."
    )] = False,
    coordinates: Annotated[bool, typer.Option(
        "-c", "--coordinates", help="Include feature coordinates in the output."
    )] = False,
    chromosomes: Annotated[bool, typer.Option(
        "-ch", "--chromosomes", help="Include chromosome/scaffold information in the output."
    )] = False,
    coding_info: Annotated[bool, typer.Option(
        "--coding-info", help="Include whether a gene codes for a protein or not."
    )] = False,
    skip_coding: Annotated[bool, typer.Option(
        "--skip-coding", help="Whether to skip coding genes."
    )] = False,
    skip_non_coding: Annotated[bool, typer.Option(
        "--skip-non-coding", help="Whether to skip non-coding genes."
    )] = False,
    gene_symbols: Annotated[bool, typer.Option(
        "--gene-symbols", help="Whether to include gene symbols in the output."
    )] = False,
    sep: Annotated[str, typer.Option(
        "-s", "--sep", help="Separator for the output file. Tab is default."
    )] = "\t",
    skip_pseudogenes: Annotated[bool, typer.Option(
        "--skip-pseudogenes", help="Whether to skip pseudogenes."
    )] = False,
    skip_transposables: Annotated[bool, typer.Option(
        "--skip-transposables", help="Whether to skip transposable elements."
    )] = False,
    quiet: Annotated[bool, typer.Option(
        "-q", "--quiet", help="Keeps terminal reporting to a minimum."
    )] = False,
):
    """
    Lists genes from an annotation file and exports them to a TSV/CSV file.
    """
    if annotation_name == "{annotation-file}":
        annotation_name = os.path.splitext(os.path.basename(annotation_file))[0]
    
    if output_file == "{annotation-name}_genes_list.tsv":
        output_file = f"{annotation_name}_genes_list.tsv"

    os.makedirs(output_dir, exist_ok=True)

    annotation = Annotation(name=annotation_name, annot_file_path=annotation_file, quiet=quiet)

    annotation.export.gene_list(
        custom_path=output_dir,
        output_file=output_file,
        lengths=lengths,
        coordinates=coordinates,
        chromosomes=chromosomes,
        coding_info=coding_info,
        skip_coding=skip_coding,
        skip_non_coding=skip_non_coding,
        sep=sep,
        skip_pseudogenes=skip_pseudogenes,
        skip_transposables=skip_transposables,
        gene_symbols=gene_symbols
    )

@app.command()
def transcripts(
    annotation_file: Annotated[str, typer.Argument(
        help="Path to the input annotation GFF/GTF file."
    )],
    annotation_name: Annotated[str, typer.Option(
        "-a", "--annotation-name", help="Annotation version, name or tag. [default: derived from filename]"
    )] = "{annotation-file}",
    output_dir: Annotated[str, typer.Option(
        "-d", "--output-dir", help="Path to the output folder."
    )] = "./aegis_output/",
    output_file: Annotated[str, typer.Option(
        "-o", "--output-file", help="Path to the output filename, without extension."
    )] = "{annotation-name}_transcripts_list.tsv",
    lengths: Annotated[bool, typer.Option(
        "-l", "--lengths", help="Include feature lengths in the output."
    )] = False,
    coordinates: Annotated[bool, typer.Option(
        "-c", "--coordinates", help="Include feature coordinates in the output."
    )] = False,
    chromosomes: Annotated[bool, typer.Option(
        "-ch", "--chromosomes", help="Include chromosome/scaffold information in the output."
    )] = False,
    coding_info: Annotated[bool, typer.Option(
        "--coding-info", help="Include whether a gene codes for a protein or not."
    )] = False,
    skip_coding: Annotated[bool, typer.Option(
        "--skip-coding", help="Whether to skip coding transcripts."
    )] = False,
    skip_non_coding: Annotated[bool, typer.Option(
        "--skip-non-coding", help="Whether to skip non-coding transcripts."
    )] = False,
    gene_symbols: Annotated[bool, typer.Option(
        "--gene-symbols", help="Whether to include gene symbols in the output."
    )] = False,
    sep: Annotated[str, typer.Option(
        "-s", "--sep", help="Separator for the output file. Default is tab."
    )] = "\t",
    skip_pseudogenes: Annotated[bool, typer.Option(
        "--skip-pseudogenes", help="Whether to skip pseudogenes."
    )] = False,
    skip_transposables: Annotated[bool, typer.Option(
        "--skip-transposables", help="Whether to skip transposable elements."
    )] = False,
    quiet: Annotated[bool, typer.Option(
        "-q", "--quiet", help="Keeps terminal reporting to a minimum."
    )] = False,
):
    """
    Lists transcripts from an annotation file and exports them to a TSV/CSV file.
    """
    if annotation_name == "{annotation-file}":
        annotation_name = os.path.splitext(os.path.basename(annotation_file))[0]
    
    if output_file == "{annotation-name}_transcripts_list.tsv":
        output_file = f"{annotation_name}_transcripts_list.tsv"

    os.makedirs(output_dir, exist_ok=True)

    annotation = Annotation(name=annotation_name, annot_file_path=annotation_file, quiet=quiet)

    annotation.export.transcript_list(
        custom_path=output_dir,
        output_file=output_file,
        lengths=lengths,
        coordinates=coordinates,
        chromosomes=chromosomes,
        coding_info=coding_info,
        skip_coding=skip_coding,
        skip_non_coding=skip_non_coding,
        sep=sep,
        skip_pseudogenes=skip_pseudogenes,
        skip_transposables=skip_transposables,
        gene_symbols=gene_symbols
    )

if __name__ == "__main__":
    app()



"""
Created on Thu Jan 19 15:26:49 2023

@authors: David Navarro, Antonio Santiago
"""

import pickle
import re
import subprocess

from pathlib import Path
from collections import Counter

def pickle_load(file):
    f = open(file, "rb")
    item = pickle.load(f)
    f.close()
        
    return item

def pickle_save(file, item):
    f = open(file, "wb")
    pickle.dump(item, f)
    f.close()

def count_occurrences(string, char):
    return Counter(string)[char]

def find_all_occurrences(pattern, text):
    matches = []
    for match in re.finditer(pattern, text):
        matches.append((match.start(), match.end(), match.group()))

    return matches

def run_command(working_directory: Path, command: list):
    """
    Executes a generic command inside a Docker container.

    Args:
        working_directory (Path): The working directory for the command.
        command (list): The command and its arguments as a list of strings.

    Raises:
        subprocess.CalledProcessError: If the command fails.
    """
    try:
        result = subprocess.run(command, check=True, capture_output=True, text=True, cwd=working_directory)
        return result
    except subprocess.CalledProcessError as e:
        print(f"Error executing command: {' '.join(command)}")
        print(f"STDOUT: {e.stdout}")
        print(f"STDERR: {e.stderr}")
        raise


def read_file_with_fallback(file_path, encodings=['utf-8', 'latin-1', 'ascii']):
    """
    Tries several encodings to find the suitable one.
    """
    for enc in encodings:
        try:
            with open(file_path, 'r', encoding=enc) as f:
                f.readlines()
                return enc
        except UnicodeDecodeError:
            continue

    raise ValueError(f"Not able to decodify '{file_path}'")


import typer
import os

from typing_extensions import Annotated

from ..annotation import Annotation, read_file_with_fallback

features = ["gene", "transcript"]

app = typer.Typer(add_completion=False, no_args_is_help=True)
@app.command()
def main(
    annotation_file: Annotated[str, typer.Argument(
        help="Path to the input annotation GFF/GTF file."
    )],
    input_id_file: Annotated[str, typer.Argument(
        help="Input file with list of ids, one per line."
    )],
    annotation_name: Annotated[str, typer.Option(
        "-a", "--annotation-name", help="Annotation version, name or tag."
    )] = "{annotation-file}",
    feature_type: Annotated[str, typer.Option(
        "-f", "--feature-type", help=f"Identify feature level to be removed, based on input ids. Choose from {features}."
    )] = "gene",
    output_dir: Annotated[str, typer.Option(
        "-d", "--output-dir", help="Path to the output directory."
    )] = "./aegis_output/",
    output_file: Annotated[str, typer.Option(
        "-o", "--output-file", help="Path to the output annotation filename, without extension."
    )] = "{annotation-name}_pruned",
    quiet: Annotated[bool, typer.Option(
        "-q", "--quiet", help="Keeps terminal reporting to a minimum."
    )] = False,
):
    """
    Remove a list of ids, from a file, from the current annotation.
    """

    if annotation_name == "{annotation-file}":
        annotation_name = os.path.splitext(os.path.basename(annotation_file))[0]

    os.makedirs(output_dir, exist_ok=True)

    annotation = Annotation(name=annotation_name, annot_file_path=annotation_file, quiet=quiet)

    if output_file == "{annotation-name}.{ext}":
        output_file = f"{annotation_name}"

    if feature_type not in features:
        raise typer.BadParameter(f"Invalid feature level: {feature_type}. Choose from: {features}")

    if output_file == "{annotation-name}_pruned":
        output_file = f"{annotation_name}_pruned"

    input_ids = set()
   
    encoding = read_file_with_fallback(input_id_file)

    f_in = open(input_id_file, encoding=encoding)
    for line in f_in:
        if line.startswith("#"):
            continue
        id = line.strip()
        input_ids.add(id)
    f_in.close()

    if feature_type == "gene":
        annotation.remove_genes(input_ids, override_rescue=True, quiet=quiet)
    
    else:
        annotation.remove_transcripts(input_ids, quiet=quiet)

    output_file += ".gff3"
    annotation.export.gff(custom_path=output_dir, tag=output_file, quiet=quiet)


if __name__ == "__main__":
    app()



"""
Tests for aegis.genome — Scaffold and Genome classes.
"""

import pytest

from aegis.genome import Scaffold, Genome


# ============================================================
# Scaffold
# ============================================================

class TestScaffold:
    def test_basic_init(self):
        s = Scaffold("chr1", "ATGCGATCGATCGATCGATCGATCGATCG")
        assert s.name == "chr1"
        assert s.size == 29
        assert s.original_name == "chr1"

    def test_custom_original_name(self):
        s = Scaffold("chr1_renamed", "ATGCGATCG", original_name="scaffold_001")
        assert s.original_name == "scaffold_001"
        assert s.name == "chr1_renamed"

    def test_chromosome_detection(self):
        """Names starting with 'ch' set chromosome=True."""
        s = Scaffold("chr1", "ATGCGATCG")
        assert s.chromosome is True

    def test_non_chromosome(self):
        """Names not starting with 'ch' are not chromosomes."""
        s = Scaffold("scaffold_99", "ATGCGATCG")
        assert s.chromosome is False

    def test_mitochondria_detection(self):
        s = Scaffold("chrM", "ATGCGATCG")
        assert s.mitochondria is True

    def test_chloroplast_detection(self):
        s = Scaffold("chrC", "ATGCGATCG")
        assert s.chloroplast is True

    def test_unknown_chromosome(self):
        """The 'chrUn' name is detected as unknown_chromosome."""
        s = Scaffold("chrUn", "ATGCGATCG")
        assert s.unknown_chromosome is True

    def test_update_with_new_name(self):
        s = Scaffold("scaffold_1", "ATGCGATCG")
        s.update(new_name="chr5")
        assert s.name == "chr5"
        assert s.chromosome is True

    def test_copy(self):
        s = Scaffold("chr1", "ATGCGATCG")
        s2 = s.copy()
        s2.name = "changed"
        assert s.name == "chr1"

    def test_dapfit_numbered_chromosome(self):
        s = Scaffold("chr1", "ATGCGATCG")
        assert s.dapfit is True

    def test_dapfit_non_numbered(self):
        s = Scaffold("chrM", "ATGCGATCG")
        assert s.dapfit is False


# ============================================================
# Genome
# ============================================================

class TestGenome:
    def test_load_from_fasta(self, sample_fasta_file):
        g = Genome("test_genome", sample_fasta_file, quiet=True)
        assert g.name == "test_genome"
        assert len(g.scaffolds) == 2
        assert "chr1" in g.scaffolds
        assert "chr2" in g.scaffolds

    def test_scaffold_sizes(self, sample_fasta_file):
        g = Genome("test_genome", sample_fasta_file, quiet=True)
        assert g.scaffolds["chr1"].size > 0
        assert g.scaffolds["chr2"].size > 0

    def test_update(self, sample_fasta_file):
        """Genome.update sets self.size (total genome size)."""
        g = Genome("test_genome", sample_fasta_file, quiet=True)
        g.update()
        assert g.size > 0
        assert len(g.scaffolds) == 2

    def test_remove_features(self, sample_fasta_file):
        g = Genome("test_genome", sample_fasta_file, quiet=True)
        g.remove_features({"chr2"})
        assert "chr2" not in g.scaffolds
        assert "chr1" in g.scaffolds

    def test_subset(self, sample_fasta_file):
        g = Genome("test_genome", sample_fasta_file, quiet=True)
        g.subset(chosen_features={"chr1"})
        assert "chr1" in g.scaffolds

    def test_copy(self, sample_fasta_file):
        g = Genome("test_genome", sample_fasta_file, quiet=True)
        g2 = g.copy()
        g2.name = "changed"
        assert g.name == "test_genome"



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..annotation import Annotation
    from ..genome import Genome

import matplotlib.pyplot as plt
from scipy.stats import fisher_exact
from tqdm import tqdm
import pandas as pd
import os
import sys

from ..utils.genefunctions import reverse_complement
from ..utils.misc import find_all_occurrences

class AnnotationMotifs:
    """
    Component for handling motif methods for the Annotation class.
    Accessed via 'annotation_object.motifs'.
    """
    def __init__(self, annotation: Annotation):
        self._annot = annotation

    def find_and_plot(self, query_genes:list, motif:str, motif_length:int, glistname, tf_motif_tag, backlist:list=[], backlistname:str="", custom_path:str="", quiet:bool=False):
        # Check if stdout or stderr are redirected to files
        stdout_redirected = not sys.stdout.isatty()
        stderr_redirected = not sys.stderr.isatty()

        # Disable tqdm if stdout or stderr are redirected
        if stdout_redirected or stderr_redirected or quiet:
            disable = True
        else:
            disable = False

        bin_division = 30
        bins_genome_division = 30

        if backlist != [] and not backlistname:
            backlistname = "custom_background"

        if motif_length < 4 or len(motif) < 4:
            raise ValueError(f"Chosen motif={motif} is too short (len={motif_length}) for promoter search.")

        random_ids = self._annot.return_random_gene_ids(len(query_genes), to_avoid=query_genes)
        if backlist == []:
            total = (len(query_genes) * 2) + len(self._annot.all_gene_ids.keys())
        else:
            total = (len(query_genes) * 2) + len(backlist)
        progress_bar = tqdm(total=total, disable=disable,
                                bar_format=(
                    f'\033[1;94;1mScanning {glistname} genes for {tf_motif_tag} ({motif}):\033[0m '
                    '{percentage:3.0f}%|'
                    f'\033[1;94;1m{{bar}}\033[0m| '
                    '{n}/{total} [{elapsed}<{remaining}]'))
        
        if custom_path == "":
            output_path = self._annot.path + "motifs/"
            output_file = output_path 
        else:
            output_file = custom_path
            if output_file[-1] != "/":
                output_file += "/"
            output_file += "motifs/"

        os.makedirs(output_file, exist_ok=True)

        output_file += f"{tf_motif_tag}"

        motif = motif.upper()
        promoter_length = 0

        for genes in self._annot.chrs.values():
            for g in genes.values():
                for t in g.transcripts.values():
                    if t.main:
                        if hasattr(t, "promoter"):
                            if t.promoter.seq != "":
                                promoter_length = len(t.promoter.seq)
                                break
                break
            break
                    
        # critical thing to understand here is that we are looking at how a motif
        # is oriented with regards to the TSS
        towards_occurrences = []
        against_occurrences = []
        interest_proportion = 0
        avg_motifs_interest = []
        for id in query_genes:
            progress_bar.update(1)
            ch = self._annot.all_gene_ids[id]
            g = self._annot.chrs[ch][id]
            for t in g.transcripts.values():
                if t.main:
                    p = t.promoter
                    occurrences_t = find_all_occurrences(motif, p.seq)
                    occurrences_a = find_all_occurrences(motif, reverse_complement(p.seq))
                    occurrence_count_total = len(occurrences_t) + len(occurrences_a)
                    if occurrence_count_total != 0:
                        avg_motifs_interest.append(occurrence_count_total)
                    if occurrences_t != [] or occurrences_a != []:
                        interest_proportion += 1
                    for m in occurrences_t:
                        towards_occurrences.append((m[0], m[1], p.start + m[0], p.start + m[1], m[2], "same", g.id, str(g.names), g.strand, g.ch))
                    for m in occurrences_a:
                        against_occurrences.append((m[0], m[1], p.end - m[0], p.end - m[1], m[2], "different", g.id, str(g.names), g.strand, g.ch))


        midpoints = []
        for o in towards_occurrences:
            midpoint = (o[0] + o[1]) // 2
            midpoint -= promoter_length
            midpoints.append(midpoint)

        for o in against_occurrences:
            midpoint = (o[0] + o[1]) // 2
            midpoint += 1
            midpoint = (midpoint * -1)
            midpoints.append(midpoint)


        all_occurrences = towards_occurrences + against_occurrences
        df = pd.DataFrame(all_occurrences, columns=["start", "end", "genomic_start", "genomic_end", "sequence", "orientation with respect to gene", "gene_id", "gene_name", "gene_strand", "chromosome"])
        if backlist == []:
            df.to_csv(f"{output_file}_{glistname}.csv", sep="\t")
        
        interest_count = len(midpoints)
        plt.hist(midpoints, bins=(promoter_length//bin_division), color='skyblue', edgecolor='skyblue')
        plt.grid(False)
        plt.title(f"{self._annot.id} {tf_motif_tag} {glistname} histogram\npromoters with motif: ({interest_proportion}/{len(query_genes)})")
        plt.xlabel("promoter position")
        plt.ylabel(f"motif occurrence count (total: {interest_count})")
        plt.grid(True)
        if backlist == []:
            plt.savefig(f"{output_file}_{glistname}.pdf")
        plt.close()

        # critical thing to understand here is that we are looking at how a motif
        # is oriented with regards to the TSS
        towards_occurrences = []
        against_occurrences = []
        avg_motifs_random = []
        random_proportion = 0
        for id in random_ids:
            progress_bar.update(1)
            ch = self._annot.all_gene_ids[id]
            g = self._annot.chrs[ch][id]
            for t in g.transcripts.values():
                if t.main:
                    p = t.promoter
                    occurrences_t = find_all_occurrences(motif, p.seq)
                    occurrences_a = find_all_occurrences(motif, reverse_complement(p.seq))
                    occurrence_count_total = len(occurrences_t) + len(occurrences_a)
                    if occurrence_count_total != 0:
                        avg_motifs_random.append(occurrence_count_total)
                    if occurrences_t != [] or occurrences_a != []:
                        random_proportion += 1
                    for m in occurrences_t:
                        towards_occurrences.append((m[0], m[1], p.start + m[0], p.start + m[1], m[2], "same", g.id, str(g.names), g.strand, g.ch))
                    for m in occurrences_a:
                        against_occurrences.append((m[0], m[1], p.end - m[0], p.end - m[1], m[2], "different", g.id, str(g.names), g.strand, g.ch))

        midpoints = []
        for o in towards_occurrences:
            midpoint = (o[0] + o[1]) // 2
            midpoint -= promoter_length
            midpoints.append(midpoint)

        for o in against_occurrences:
            midpoint = (o[0] + o[1]) // 2
            midpoint += 1
            midpoint = (midpoint * -1)
            midpoints.append(midpoint)


        all_occurrences = towards_occurrences + against_occurrences
        df = pd.DataFrame(all_occurrences, columns=["start", "end", "genomic_start", "genomic_end", "sequence", "orientation with respect to gene", "gene_id", "gene_name", "gene_strand", "chromosome"])
        if backlist == []:
            df.to_csv(f"{output_file}_{glistname}_random.csv", sep="\t")

        random_count = len(midpoints)
        plt.hist(midpoints, bins=(promoter_length//bin_division), color='grey', edgecolor='grey')
        plt.grid(False)
        plt.title(f"{self._annot.id} {tf_motif_tag} {glistname} random histogram\npromoters with motif: ({random_proportion}/{len(query_genes)})")
        plt.xlabel("promoter position")
        plt.ylabel(f"motif occurrence count (total: {random_count})")
        plt.grid(True)
        if backlist == []:
            plt.savefig(f"{output_file}_{glistname}_random.pdf")
        plt.close()

        # critical thing to understand here is that we are looking at how a motif
        # is oriented with regards to the TSS
        towards_occurrences = []
        against_occurrences = []
        avg_motifs_genomic = []
        genomic_proportion = 0
        if backlist == []:
            for id in self._annot.all_gene_ids:
                progress_bar.update(1)
                ch = self._annot.all_gene_ids[id]
                g = self._annot.chrs[ch][id]
                for t in g.transcripts.values():
                    if t.main:
                        p = t.promoter
                        occurrences_t = find_all_occurrences(motif, p.seq)
                        occurrences_a = find_all_occurrences(motif, reverse_complement(p.seq))
                        occurrence_count_total = len(occurrences_t) + len(occurrences_a)
                        if occurrence_count_total != 0:
                            avg_motifs_genomic.append(occurrence_count_total)
                        if occurrences_t != [] or occurrences_a != []:
                            genomic_proportion += 1
                        for m in occurrences_t:
                            towards_occurrences.append((m[0], m[1], p.start + m[0], p.start + m[1], m[2], "same", g.id, str(g.names), g.strand, g.ch))
                        for m in occurrences_a:
                            against_occurrences.append((m[0], m[1], p.end - m[0], p.end - m[1], m[2], "different", g.id, str(g.names), g.strand, g.ch))

            midpoints = []
            for o in towards_occurrences:
                midpoint = (o[0] + o[1]) // 2
                midpoint -= promoter_length
                midpoints.append(midpoint)

            for o in against_occurrences:
                midpoint = (o[0] + o[1]) // 2
                midpoint += 1
                midpoint = (midpoint * -1)
                midpoints.append(midpoint)

            genomic_count = len(midpoints)
            plt.hist(midpoints, bins=(promoter_length//bins_genome_division), color='grey', edgecolor='grey')
            plt.grid(False)
            plt.title(f"{self._annot.id} {tf_motif_tag} full genome histogram\npromoters with motif: ({genomic_proportion}/{len(self._annot.all_gene_ids.keys())})")
            plt.xlabel("promoter position")
            plt.ylabel(f"motif occurrence count (total: {genomic_count})")
            plt.grid(True)
            plt.savefig(f"{output_file}_whole_genome.pdf")
            plt.close()
        
        else:
            for id in backlist:
                progress_bar.update(1)
                ch = self._annot.all_gene_ids[id]
                g = self._annot.chrs[ch][id]
                for t in g.transcripts.values():
                    if t.main:
                        p = t.promoter
                        occurrences_t = find_all_occurrences(motif, p.seq)
                        occurrences_a = find_all_occurrences(motif, reverse_complement(p.seq))
                        occurrence_count_total = len(occurrences_t) + len(occurrences_a)
                        if occurrence_count_total != 0:
                            avg_motifs_genomic.append(occurrence_count_total)
                        if occurrences_t != [] or occurrences_a != []:
                            genomic_proportion += 1
                        for m in occurrences_t:
                            towards_occurrences.append((m[0], m[1], p.start + m[0], p.start + m[1], m[2], "same", g.id, str(g.names), g.strand, g.ch))
                        for m in occurrences_a:
                            against_occurrences.append((m[0], m[1], p.end - m[0], p.end - m[1], m[2], "different", g.id, str(g.names), g.strand, g.ch))

            midpoints = []
            for o in towards_occurrences:
                midpoint = (o[0] + o[1]) // 2
                midpoint -= promoter_length
                midpoints.append(midpoint)

            for o in against_occurrences:
                midpoint = (o[0] + o[1]) // 2
                midpoint += 1
                midpoint = (midpoint * -1)
                midpoints.append(midpoint)

            genomic_count = len(midpoints)
            plt.hist(midpoints, bins=(promoter_length//bins_genome_division), color='grey', edgecolor='grey')
            plt.grid(False)
            plt.title(f"{self._annot.id} {tf_motif_tag} {backlistname} as background histogram\npromoters with motif: ({genomic_proportion}/{len(backlist)})")
            plt.xlabel("promoter position")
            plt.ylabel(f"motif occurrence count (total: {genomic_count})")
            plt.grid(True)
            plt.savefig(f"{output_file}_{backlistname}_as_background.pdf")
            plt.close()

        progress_bar.close()

        # Counts of non-motif occurrences
        interest_non_count = (len(query_genes) * (int(promoter_length/motif_length)) * 2) - interest_count
        if backlist == []:
            genomic_non_count = (len(self._annot.all_gene_ids.keys()) * (int(promoter_length/motif_length)) * 2) - genomic_count
        else:
            genomic_non_count = (len(backlist) * (int(promoter_length/motif_length)) * 2) - genomic_count
        interest_non_proportion = len(query_genes) - interest_proportion
        if backlist == []:
            genomic_non_proportion = len(self._annot.all_gene_ids.keys()) - genomic_proportion
        else:
            genomic_non_proportion = len(backlist) - genomic_proportion

        odds_ratio_occurrences, p_value_occurrences = fisher_exact([[interest_count, genomic_count], 
                                                                    [interest_non_count, genomic_non_count]])

        odds_ratio_proportion, p_value_proportion = fisher_exact([[interest_proportion, genomic_proportion], 
                                                                  [interest_non_proportion, genomic_non_proportion]])

        promoter_percentage_interest = (interest_proportion / len(query_genes)) * 100
        if backlist == []:
            promoter_percentage_genome = (genomic_proportion / len(self._annot.all_gene_ids.keys())) * 100
        else:
            promoter_percentage_genome = (genomic_proportion / len(backlist)) * 100

        return interest_count, genomic_count, p_value_occurrences, odds_ratio_occurrences, promoter_percentage_interest, promoter_percentage_genome, p_value_proportion, odds_ratio_proportion, avg_motifs_interest, avg_motifs_genomic, output_file



from .annotation import Annotation
from .feature import Feature
from .gene import Gene
from .genome import Genome, Scaffold
from .transcript import Transcript
from .subfeatures import CDS, Exon, UTR, Intron
from .misc_features import Protein, Promoter
from .hits import OverlapHit, BlastHit


from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .genome import Genome

import pandas as pd
import re
import time
import os
import shutil

from pathlib import Path

from .annotation import Annotation
from .utils.misc import run_command
from .utils.evalue import parse_evalue, round_evalue

def pairwise_orthology(annot1: Annotation, annot2: Annotation, genome1: Genome, genome2: Genome, working_directory: Path, num_threads: int, types: str, evalue:float=0.00001, coverage:float=30, max_hsps:int=1, copies:bool=True, synteny:bool=False, skip_lifton:bool=False, skip_mcscan:bool=False, quiet:bool=True):

    liftoff_dir = working_directory / "liftoff"
    lifton_dir = working_directory / "lifton"
    protein_dir = working_directory / "proteins"
    diamond_dir = working_directory / "diamond"
    mcscan_dir = working_directory / "mcscan"

    print(f"\n\n{annot1.name} vs {annot2.name}:")

    print(f"\n\tRunning Liftoff to map annotations from {annot1.name} on {annot2.name}")

    liftoff_gff = liftoff_dir / f"liftoff_{annot1.name}_to_{annot2.name}.gff"
    liftoff_cmd = [
        "liftoff", str(genome2.file), str(genome1.file),
        "-g", f"{working_directory}/gffs/{annot1.name}.gff3", "-o", str(liftoff_gff), "-flank",  "0.1", "-f", types
    ]
    if copies:
        liftoff_cmd.append("-copies")

    run_command(liftoff_dir, liftoff_cmd)

    to_remove = liftoff_dir / "intermediate_files"

    if os.path.exists(str(to_remove)):
        shutil.rmtree(str(to_remove))
        unmapped_file = f"{str(liftoff_dir)}/unmapped_features.txt"
        if os.path.isfile(unmapped_file):
            os.remove(unmapped_file)

    print(f"\t\tRunning aegis-overlaps on liftoff result.")

    if synteny:
        a_liftoff = Annotation(str(liftoff_gff), original_annotation=annot1, quiet=quiet)
    else:
        a_liftoff = Annotation(str(liftoff_gff), quiet=quiet)

    a_liftoff.overlaps.detect(annot2, quiet=quiet)

    _ = a_liftoff.overlaps.export(custom_path=str(liftoff_dir), output_file=f"liftoff_{annot1.name}_to_{annot2.name}_overlaps.tsv", verbose=True, export_csv=True, NAs=False, quiet=quiet, synteny=synteny, copies_info=True)

    del a_liftoff

    if not skip_lifton:

        print(f"\n\tRunning Lifton to map annotations from {annot1.name} on {annot2.name}")

        lifton_gff = lifton_dir / f"lifton_{annot1.name}_to_{annot2.name}.gff3"
        lifton_cmd = [
            "lifton", "-g", f"{working_directory}/gffs/{annot1.name}_for_lifton.gff3", "-o", str(lifton_gff),
            "-flank",  "0.1", "-f", types
        ]
        if copies:
            lifton_cmd.append("-copies")

        lifton_cmd.append(str(genome2.file))
        lifton_cmd.append(str(genome1.file))

        run_command(lifton_dir, lifton_cmd)

        to_remove = lifton_dir / "lifton_output"

        if os.path.exists(str(to_remove)):
            shutil.rmtree(str(to_remove))

        if not os.path.isfile(lifton_gff):
            with open(lifton_gff, "w") as f:
                f.write("##gff-version 3\n")

        print(f"\t\tRunning aegis-overlaps on lifton result.")

        if synteny:
            a_lifton = Annotation(str(lifton_gff), original_annotation=annot1, quiet=quiet)
        else:
            a_lifton = Annotation(str(lifton_gff), quiet=quiet)

        a_lifton.overlaps.detect(annot2, quiet=quiet)

        _ = a_lifton.overlaps.export(custom_path=str(lifton_dir), output_file=f"lifton_{annot1.name}_to_{annot2.name}_overlaps.tsv", verbose=True, export_csv=True, NAs=False, quiet=quiet, synteny=synteny, copies_info=True)

        del a_lifton

    protein_fasta = protein_dir / f"{annot1.name}_proteins_g_id_main.fasta"

    diamond_db = diamond_dir / f"{annot2.name}_diamond_db"

    diamond_result = diamond_dir / f"single_{annot1.name}_to_{annot2.name}.txt"
    diamond_result_best = diamond_dir / f"single_best_{annot1.name}_to_{annot2.name}.txt"

    print(f"\n\tRunning DIAMOND search ({annot1.name} -> {annot2.name})")
    blastp_cmd = [
        "diamond", "blastp", "--threads", str(num_threads), "--db", str(diamond_db), "--ultra-sensitive", 
        "--out", str(diamond_result), "--outfmt", "6", "qseqid", "sseqid", "pident", "qcovhsp", 
        "qlen", "slen", "length", "bitscore", "evalue", "--query", str(protein_fasta), 
        "--evalue", str(evalue), "--max-hsps", str(max_hsps), "--query-cover", str(coverage)
    ]

    run_command(diamond_dir, blastp_cmd)

    blastp_cmd = [
        "diamond", "blastp", "--threads", str(num_threads), "--db", str(diamond_db), "--ultra-sensitive", 
        "--out", str(diamond_result_best), "--outfmt", "6", "qseqid", "sseqid", "pident", "qcovhsp", 
        "qlen", "slen", "length", "bitscore", "evalue", "--query", str(protein_fasta), 
        "--max-target-seqs", "1", "--evalue", str(evalue), "--max-hsps", str(max_hsps)
    ]

    run_command(diamond_dir, blastp_cmd)
    
    if not skip_mcscan:
        print(f"\n\tRunning JCVI ortholog analysis (this may take a while) between {annot1.name} and {annot2.name}")
        
        jcvi_ortho_cmd = [
            "python", "-m", "jcvi.compara.catalog", "ortholog",
            annot1.name, annot2.name, "--no_strip_names"
        ]
        run_command(mcscan_dir, jcvi_ortho_cmd)


class Equivalence():

    preferred_type_order = ["rec_liftoff_aegis", "rec_lifton_aegis", "fwd_liftoff_aegis", "rev_liftoff_aegis", "rev_lifton_aegis", "rev_lifton_aegis", "mcscan_anchors", "mcscan_last_filtered", "rbbh", "rbh", "orthofinder", "fwd_blastp", "rev_blastp", "fwd_blast", "rev_blast"]
    reliability_order = ["vvvtop_reliable", "vvtop_reliable", "vtop_reliable", "top_reliable", "vvvvv_reliable", "vvvv_reliable", "vvv_reliable", "vv_reliable", "v_reliable", "reliable", "NA"]

    def __init__(self, id_, type_, target_annotation, species, score:str="", evalue:str|None=None, reliability:str="NA"):
        self.id = id_
        self.type = type_
        self.species = species
        self.score = score
        self.target_annotation = target_annotation
        self.reliability = reliability

        self.aegis_score = None
        self.aegis_synteny = None
        self.aegis_copies = None
        self.match_num = None

        self.aegis_score_rev = None
        self.aegis_synteny_rev = None
        self.aegis_copies_rev = None
        self.match_num_rev = None

        self.coverage = None
        self.identity = None

        self.coverage_rev = None
        self.identity_rev = None

        if evalue != None:
            self.evalue = parse_evalue(evalue)
        else:
            self.evalue = None


        if "confidence" not in self.score:

            if "blast" in self.type or "rbh" in self.type or "rbbh" in self.type:
                for score_component in self.score.split(", "):
                    if score_component.startswith("coverage="):
                        self.identity = round(float(score_component.split("=")[-1].split("/")[0]), 1)
                    elif score_component.startswith("identity="):
                        self.coverage = round(float(score_component.split("=")[-1].split("/")[0]), 1)

                if "rbh" in self.type or "rbbh" in self.type:
                    for score_component in self.score.split(", "):
                        if score_component.startswith("coverage="):
                            self.identity_rev = round(float(score_component.split("=")[-1].split("/")[1]), 1)
                        elif score_component.startswith("identity="):
                            self.coverage_rev = round(float(score_component.split("=")[-1].split("/")[1]), 1)
                
            elif "aegis" in self.type:

                scores = self.score.split("/")

                self.aegis_score = int(scores[0].split(" (")[0])

                if "copies" in scores[0]:
                    self.aegis_copies = True
                else:
                    self.aegis_copies = False

                if "multiple" in scores[0]:
                    self.match_num = int(scores[0].split("multiple ")[-1].split(")")[0])
                else:
                    self.match_num = 1            

                if "synteny" in scores[0]:
                    self.aegis_synteny = True
                else:
                    self.aegis_synteny = False

                if "rec_" in self.type:

                    self.aegis_score_rev = int(scores[1].split(" (")[0])

                    if "copies" in scores[1]:
                        self.aegis_copies_rev = True
                    else:
                        self.aegis_copies_rev = False

                    if "multiple" in scores[1]:
                        self.match_num_rev = int(scores[1].split("multiple ")[-1].split(")")[0])
                    else:
                        self.match_num_rev = 1            

                    if "synteny" in scores[1]:
                        self.aegis_synteny_rev = True
                    else:
                        self.aegis_synteny_rev = False

        else:
            for master_score_component in self.type.split("], "):
                if master_score_component.endswith("]"):
                    master_score_component = master_score_component[:-1]

                if "blast" in master_score_component or "rbh" in master_score_component or "rbbh" in master_score_component:
                    for score_component in master_score_component.split("[")[1].split(", "):
                        if score_component.startswith("coverage="):
                            self.identity = round(float(score_component.split("=")[-1].split("/")[0]), 1)
                        elif score_component.startswith("identity="):
                            self.coverage = round(float(score_component.split("=")[-1].split("/")[0]), 1)

                    if "rbh" in master_score_component or "rbbh" in master_score_component:
                        for score_component in master_score_component.split("[")[1].split(", "):
                            if score_component.startswith("coverage="):
                                self.identity_rev = round(float(score_component.split("=")[-1].split("/")[1]), 1)
                            elif score_component.startswith("identity="):
                                self.coverage_rev = round(float(score_component.split("=")[-1].split("/")[1]), 1)

                elif "aegis" in master_score_component:

                    if "liftoff" in self.type and "lifton" in master_score_component:
                        continue

                    scores = master_score_component.split("[")[1].split("/")

                    self.aegis_score = int(scores[0].split(" (")[0])

                    if "copies" in scores[0]:
                        self.aegis_copies = True
                    else:
                        self.aegis_copies = False

                    if "multiple" in scores[0]:
                        self.match_num = int(scores[0].split("multiple ")[-1].split(")")[0])
                    else:
                        self.match_num = 1            

                    if "synteny" in scores[0]:
                        self.aegis_synteny = True
                    else:
                        self.aegis_synteny = False

                    if "rec_" in master_score_component:

                        self.aegis_score_rev = int(scores[1].split(" (")[0])

                        if "copies" in scores[1]:
                            self.aegis_copies_rev = True
                        else:
                            self.aegis_copies_rev = False

                        if "multiple" in scores[1]:
                            self.match_num_rev = int(scores[1].split("multiple ")[-1].split(")")[0])
                        else:
                            self.match_num_rev = 1            

                        if "synteny" in scores[1]:
                            self.aegis_synteny_rev = True
                        else:
                            self.aegis_synteny_rev = False

    def _rank(self):
        if self.reliability == "NA":
            target_rank = self.target_annotation.lower()
            type_rank = self.preferred_type_order.index(self.type) if self.type in self.preferred_type_order else float('inf')
            evalue = self.evalue if isinstance(self.evalue, float) else float('inf')
            aegis_score = self.aegis_score if self.aegis_score != None else 100
            aegis_score_rev = self.aegis_score_rev if self.aegis_score_rev != None else 100

            if self.aegis_synteny == None:
                synteny = 3
            elif self.aegis_synteny:
                synteny = 1
            else:
                synteny = 2

            if self.aegis_synteny_rev == None:
                synteny_rev = 3
            elif self.aegis_synteny_rev:
                synteny_rev = 1
            else:
                synteny_rev = 2

            if self.aegis_copies == None:
                copies = 3
            elif self.aegis_copies:
                copies = 1
            else:
                copies = 2

            if self.aegis_copies_rev == None:
                copies_rev = 3
            elif self.aegis_copies_rev:
                copies_rev = 1
            else:
                copies_rev = 2

            return (target_rank, type_rank, evalue, -aegis_score, -aegis_score_rev, synteny, synteny_rev, copies, copies_rev)
        else:
            target_rank = self.target_annotation.lower()
            reliability_rank = self.reliability_order.index(self.reliability) if self.reliability in self.reliability_order else float('inf')
            evalue = self.evalue if isinstance(self.evalue, float) else float('inf')
            aegis_score = self.aegis_score if self.aegis_score != None else 100
            aegis_score_rev = self.aegis_score_rev if self.aegis_score_rev != None else 100

            if self.aegis_synteny == None:
                synteny = 3
            elif self.aegis_synteny:
                synteny = 1
            else:
                synteny = 2

            if self.aegis_synteny_rev == None:
                synteny_rev = 3
            elif self.aegis_synteny_rev:
                synteny_rev = 1
            else:
                synteny_rev = 2

            if self.aegis_copies == None:
                copies = 3
            elif self.aegis_copies:
                copies = 1
            else:
                copies = 2

            if self.aegis_copies_rev == None:
                copies_rev = 3
            elif self.aegis_copies_rev:
                copies_rev = 1
            else:
                copies_rev = 2
            
            return (target_rank, reliability_rank, evalue, -aegis_score, -aegis_score_rev, synteny, synteny_rev, copies, copies_rev)

    def __lt__(self, other):
        return self._rank() < other._rank()

    def __eq__(self, other):
        return self._rank() == other._rank()

    def __str__(self):
        return f"{self.id}\t{self.type}\t{self.score}\t{self.target_annotation}\t{self.species}"

    def __repr__(self):
        return f"{self.id}\t{self.type}\t{self.score}\t{self.target_annotation}\t{self.species}"

    def verbose(self):
        return f"{self.id}\t{self.type}\t{self.score}\t{self.target_annotation}\t{self.species}\t{self.reliability}"

class Simple_gene():
    def __init__(self, id):
        self.id = id
        self.equivalences = []
        self.filtered_equivalences = []

    def filter_equivalences(self, simple_rbh_blasts:bool=True, unidirectional_blasts:bool=True, replace:bool=True, identity_threshold=0, coverage_threshold=0, evalue_threshold=float('inf')):
        """
        For now thresholds only apply to rbhs and unidirectional_blasts.
        """

        if self.filtered_equivalences == [] or replace:

            self.filtered_equivalences = []

            equivalent_genes_temp_d = {}

            for equivalence in self.equivalences:
                if equivalence.target_annotation not in equivalent_genes_temp_d:
                    equivalent_genes_temp_d[equivalence.target_annotation] = []
            
            for equivalence in self.equivalences:
                if equivalence.id not in equivalent_genes_temp_d[equivalence.target_annotation]:
                    equivalent_genes_temp_d[equivalence.target_annotation].append(equivalence.id)

            for target_annotation, equivalent_genes in equivalent_genes_temp_d.items():
                for equivalent_gene in equivalent_genes:

                    evidences = []
                    liftoff_score = 0
                    lifton_score = 0

                    lowest_evalue = float('inf')
                    found_evalue = False
                    species = ""

                    for equivalence in self.equivalences:
                        if equivalence.id == equivalent_gene:
                            species = equivalence.species
                            if "aegis" in equivalence.type:
                                if "rec_" in equivalence.type:
                                    score = max(equivalence.aegis_score, equivalence.aegis_score_rev)
                                else:
                                    score = equivalence.aegis_score
                                if "lifton" in equivalence.type:
                                    if score > lifton_score:
                                        lifton_score = score
                                else:
                                    if score > liftoff_score:
                                        liftoff_score = score
                            elif "rbbh" in equivalence.type or "rbh" in equivalence.type or "blast" in equivalence.type:

                                if not unidirectional_blasts and "blast" in equivalence.type:
                                    continue

                                if not simple_rbh_blasts and "rbh" in equivalence.type:
                                    continue

                                if "blast" in equivalence.type or "rbh" in equivalence.type:
                                    if equivalence.identity < identity_threshold:
                                        continue
                                    if equivalence.coverage < coverage_threshold:
                                        continue
                                    if equivalence.evalue > evalue_threshold:
                                        continue

                                if equivalence.evalue < lowest_evalue:
                                    lowest_evalue = equivalence.evalue
                                found_evalue = True

                            evidences.append(f"{equivalence.type} [{equivalence.score}]")

                    summary_evidence = ", ".join(evidences)

                    if (liftoff_score >= 9 or lifton_score >= 9) and "rbbh" in summary_evidence and "mcscan_anchors" in summary_evidence:
                        reliability = "vvvtop_reliable"
                    elif liftoff_score >= 9 or lifton_score >= 9:
                        reliability = "vvtop_reliable"

                    elif liftoff_score >= 7 or lifton_score >= 7 and "rbbh" in summary_evidence and "mcscan_anchors" in summary_evidence:
                        reliability = "vtop_reliable"
                    elif liftoff_score >= 7 or lifton_score >= 7:
                        reliability = "top_reliable"

                    elif liftoff_score >= 6 or lifton_score >= 6 and "rbbh" in summary_evidence and "mcscan_anchors" in summary_evidence:
                        reliability = "vvvvv_reliable"
                    elif "rbbh" in summary_evidence or (liftoff_score >= 6 or lifton_score >= 6 and "rbh" in summary_evidence and "mcscan_anchors" in summary_evidence):
                        reliability = "vvvv_reliable"
                    elif liftoff_score >= 6 or lifton_score >= 6 and "rbh" in summary_evidence in summary_evidence:
                        reliability = "vvv_reliable"

                    elif "mcscan_anchors" in summary_evidence and "rbh" in summary_evidence:
                        reliability = "vv_reliable"

                    elif "mcscan_anchors" in summary_evidence or "rbh" in summary_evidence:
                        reliability = "v_reliable"                   

                    else:
                        reliability = "reliable"

                    if found_evalue:
                        evalue = str(lowest_evalue)
                    else:
                        evalue = "NA"

                    if "top_reliable" in reliability:
                        score = "high_confidence"
                    elif "vv_reliable" in reliability:
                        score = "medium_confidence"
                    else:
                        score = "lower_confidence"

                    if evidences != []:

                        self.filtered_equivalences.append(Equivalence(equivalent_gene, summary_evidence, target_annotation, species, score, evalue, reliability))


class Simple_annotation():
    def __init__(self, name, annotation_object:Annotation, species:str):
        self.name = name
        self.genes = {}
        self.species = species
        self.target = annotation_object.target
        for gene in annotation_object.all_gene_ids:
            if gene not in self.genes:
                self.genes[gene] = Simple_gene(gene)
            else:
                print(f"Warning: repeated gene id: {gene}")

        self.added_equivalences = {}

    def export_summary_equivalences(self, output_file, filtered:bool=False, simple_rbh_blasts:bool=True, unidirectional_blasts:bool=True, replace:bool=True, identity_threshold=0, coverage_threshold=0, evalue_threshold=float('inf'), verbose:bool=True, quiet:bool=False, export_csv:bool=True, return_df:bool=False):
        
        start = time.time()
        
        out = ["gene_id_A\tgene_id_B\tscore\tsummary_score\tannotation_A\tannotation_B\tspecies_A\tspecies_B"]
        if verbose:
            out[0] += "\treliability"

        for gene in self.genes.values():
            gene.equivalences = sorted(gene.equivalences)
        
        if filtered:
            self.filter_equivalences(simple_rbh_blasts=simple_rbh_blasts, unidirectional_blasts=unidirectional_blasts, replace=replace, identity_threshold=identity_threshold, coverage_threshold=coverage_threshold, evalue_threshold=evalue_threshold)

            for gene in self.genes.values():
                gene.filtered_equivalences = sorted(gene.filtered_equivalences)
                for equivalence in gene.filtered_equivalences:
                    if verbose:
                        out.append(f"{gene.id}\t{equivalence.id}\t{equivalence.type}\t{equivalence.score}\t{self.name}\t{equivalence.target_annotation}\t{self.species}\t{equivalence.species}\t{equivalence.reliability}")
                    else:
                        out.append(f"{gene.id}\t{equivalence.id}\t{equivalence.type}\t{equivalence.score}\t{self.name}\t{equivalence.target_annotation}\t{self.species}\t{equivalence.species}")
        else:
            for gene in self.genes.values():
                for equivalence in gene.equivalences:
                    if verbose:
                        out.append(f"{gene.id}\t{equivalence.id}\t{equivalence.type}\t{equivalence.score}\t{self.name}\t{equivalence.target_annotation}\t{self.species}\t{equivalence.species}\t{equivalence.reliability}")
                    else:
                        out.append(f"{gene.id}\t{equivalence.id}\t{equivalence.type}\t{equivalence.score}\t{self.name}\t{equivalence.target_annotation}\t{self.species}\t{equivalence.species}")

        if return_df:
            columns = out[0].split('\t')
            data = [line.split('\t') for line in out[1:]]
            df = pd.DataFrame(data, columns=columns, dtype=str)
        
        if export_csv:
            out = "\n".join(out)
            f_out = open(output_file, "w", encoding="utf-8")
            f_out.write(out)
            f_out.close()

        end = time.time()
        lapse = end - start
        if not quiet:
            print(f"Exporting all filtered={filtered} equivalences for {self.name} took {round(lapse/60, 2)} minutes\n")

        if return_df:
            return df
        


    def add_mcscan_equivalences(self, file, key_col, target_annotation, species):
        """
        Adds MCScan equivalences from ".anchors" or ".last.filtered"
        """

        if os.path.isfile(file):

            df = pd.read_csv(file, comment='#', header=None, sep="\t", dtype=str)

            if key_col == "0":
                target_col = "1"
            else:
                target_col = "0"

            if file.endswith(".filtered"):
                equivalence_type = "mcscan_last_filtered"
                df.columns = ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11"]
                score_col = "11"
            else:
                equivalence_type = "mcscan_anchors"
                df.columns = ["0", "1", "2"]
                score_col = "2"

            for index, row in df.iterrows():
                gene_query = row[key_col]
                gene_target = row[target_col]
                score = row[score_col]
                self.genes[gene_query].equivalences.append(Equivalence(gene_target, equivalence_type, target_annotation, species, score))


    def add_orthofinder_equivalences(self, file, target_annotation, species):

        if os.path.isfile(file):
            df = pd.read_csv(file, sep="\t", dtype=str)
            df.columns = ["orthogroup", "query", "target"]
            for index, row in df.iterrows():
                gene_queries = row["query"].split(", ")
                gene_targets = row["target"].split(", ")
                score = row["orthogroup"]
                for gene_query in gene_queries:
                    for gene_target in gene_targets:
                        self.genes[gene_query].equivalences.append(Equivalence(gene_target, "orthofinder", target_annotation, species, score))

    def add_blast_equivalences(self, blast_folder, query_annotation, target_annotation, species, skip_rbhs:bool=False, skip_unidirectional_blasts:bool=False, proteins:bool=True, quiet:bool=False):

        if not skip_unidirectional_blasts and skip_rbhs:
            print("Warning: It makes no sense to skip RBHs when unidirectional blasts are not being skipped.")

        else:

            start = time.time()

            query_col = "query_x"
            target_col = "subject_x"
            query_col_evalue = "E-value_x"
            target_col_evalue = "E-value_y"
            query_col_normbit = "norm_bitscore_x"
            target_col_normbit = "norm_bitscore_y"
            query_col_coverage = "coverage_x"
            target_col_coverage = "coverage_y"
            query_col_identity = "identity_x"
            target_col_identity = "identity_y"

            rbbh_file = f"{blast_folder}/rbbh_{query_annotation}_to_{target_annotation}.txt"
            rbh_file = f"{blast_folder}/rbh_{query_annotation}_to_{target_annotation}.txt"
            if not os.path.isfile(rbbh_file):
                query_col = "subject_x"
                target_col = "query_x"
                query_col_evalue = "E-value_y"
                target_col_evalue = "E-value_x"
                query_col_normbit = "norm_bitscore_y"
                target_col_normbit = "norm_bitscore_x"
                query_col_coverage = "coverage_y"
                target_col_coverage = "coverage_x"
                query_col_identity = "identity_y"
                target_col_identity = "identity_x"

                rbbh_file = f"{blast_folder}/rbbh_{target_annotation}_to_{query_annotation}.txt"
                rbh_file = f"{blast_folder}/rbh_{target_annotation}_to_{query_annotation}.txt"

            fwd_file = f"{blast_folder}/single_{query_annotation}_to_{target_annotation}.txt"
            rev_file = f"{blast_folder}/single_{target_annotation}_to_{query_annotation}.txt"

            rbbh_hits = set()

            target_annotation = clean_annotation_tag(target_annotation)

            temp_start = time.time()

            if os.path.isfile(rbbh_file):
                df = pd.read_csv(rbbh_file, sep="\t", dtype=str)
                for index, row in df.iterrows():
                    gene_query = row[query_col]
                    gene_target = row[target_col]
                    evalue = row[query_col_evalue]
                    score = f"evalue={row[query_col_evalue]}/{row[target_col_evalue]}"
                    score += f", norm_bitscore={round(float(row[query_col_normbit]), 1)}/{round(float(row[target_col_normbit]), 1)}"
                    score += f", coverage={round(float(row[query_col_coverage]), 1)}/{round(float(row[target_col_coverage]), 1)}"
                    score += f", identity={round(float(row[query_col_identity]), 1)}/{round(float(row[target_col_identity]), 1)}"
                    pair = sorted([gene_query, gene_target])
                    pair = "-".join(pair)
                    rbbh_hits.add(pair)
                    self.genes[gene_query].equivalences.append(Equivalence(gene_target, "rbbh", target_annotation, species, score, evalue))

            end = time.time()
            lapse = end - temp_start
            if not quiet:
                print(f"Adding RBBH equivalences to {self.name} '{query_annotation} vs {target_annotation}' took {round(lapse/60, 2)} minutes")

            if not skip_rbhs:

                temp_start = time.time()

                rbh_hits = set()

                if os.path.isfile(rbh_file):
                    df = pd.read_csv(rbh_file, sep="\t", dtype=str)
                    for index, row in df.iterrows():
                        gene_query = row[query_col]
                        gene_target = row[target_col]
                        evalue = row[query_col_evalue]
                        score = f"evalue={row[query_col_evalue]}/{row[target_col_evalue]}"
                        score += f", norm_bitscore={round(float(row[query_col_normbit]), 1)}/{round(float(row[target_col_normbit]), 1)}"
                        score += f", coverage={round(float(row[query_col_coverage]), 1)}/{round(float(row[target_col_coverage]), 1)}"
                        score += f", identity={round(float(row[query_col_identity]), 1)}/{round(float(row[target_col_identity]), 1)}"
                        pair = sorted([gene_query, gene_target])
                        pair = "-".join(pair)
                        if pair not in rbbh_hits:
                            rbh_hits.add(pair)
                            self.genes[gene_query].equivalences.append(Equivalence(gene_target, "rbh", target_annotation, species, score, evalue))

                end = time.time()
                lapse = end - temp_start
                if not quiet:
                    print(f"Adding RBH equivalences to {self.name} '{query_annotation} vs {target_annotation}' took {round(lapse/60, 2)} minutes")

                if not skip_unidirectional_blasts:

                    temp_start = time.time()

                    if proteins:
                        equivalence_type = "fwd_blastp"
                    else:
                        equivalence_type = "fwd_blast"

                    if os.path.isfile(fwd_file):
                        df = pd.read_csv(fwd_file, sep="\t", dtype=str, header=None)
                        df.columns = ["0", "1", "2", "3", "4", "5", "6", "7", "8"]
                        for index, row in df.iterrows():
                            gene_query = row["0"]
                            gene_target = row["1"]
                            norm_bitscore = round((float(row["7"]) / float(row["4"])), 1)
                            evalue = row['8']
                            score = f"evalue={evalue}"
                            score += f", norm_bitscore={norm_bitscore}"
                            score += f", coverage={round(float(row['3']), 1)}"
                            score += f", identity={round(float(row['2']), 1)}"
                            pair = sorted([gene_query, gene_target])
                            pair = "-".join(pair)
                            if pair not in rbh_hits and pair not in rbbh_hits:
                                self.genes[gene_query].equivalences.append(Equivalence(gene_target, equivalence_type, target_annotation, species, score, evalue))

                    if proteins:
                        equivalence_type = "rev_blastp"
                    else:
                        equivalence_type = "rev_blast"

                    if os.path.isfile(rev_file):
                        df = pd.read_csv(rev_file, sep="\t", dtype=str, header=None)
                        df.columns = ["0", "1", "2", "3", "4", "5", "6", "7", "8"]
                        for index, row in df.iterrows():
                            gene_query = row["1"]
                            gene_target = row["0"]
                            norm_bitscore = round((float(row["7"]) / float(row["4"])), 1)
                            evalue = row['8']
                            score = f"evalue={evalue}"
                            score += f", norm_bitscore={norm_bitscore}"
                            score += f", coverage={round(float(row['3']), 1)}"
                            score += f", identity={round(float(row['2']), 1)}"
                            pair = sorted([gene_query, gene_target])
                            pair = "-".join(pair)
                            if pair not in rbh_hits and pair not in rbbh_hits:
                                self.genes[gene_query].equivalences.append(Equivalence(gene_target, equivalence_type, target_annotation, species, score, evalue))

                    end = time.time()
                    lapse = end - temp_start
                    if not quiet:
                        print(f"Adding fwd/rev blast equivalences to {self.name} '{query_annotation} vs {target_annotation}' took {round(lapse/60, 2)} minutes")

            end = time.time()
            lapse = end - start
            
            if not quiet:
                if skip_unidirectional_blasts and not skip_rbhs:
                    print(f"Adding RBBH, and RBH equivalences to {self.name} '{query_annotation} vs {target_annotation}' took {round(lapse/60, 2)} minutes")
                else:
                    print(f"Adding RBBH, RBH, and fwd/rev blast equivalences to {self.name} '{query_annotation} vs {target_annotation}' took {round(lapse/60, 2)} minutes")

    def add_reciprocal_overlap_equivalences(self, folder, query_tag, target_tag, species, liftoff:bool=True, quiet:bool=False, synteny_present:bool=False):

        program = "liftoff"
        if not liftoff:
            program = "lifton"

        fwd_file = f"{folder}/{program}_{query_tag}_to_{target_tag}_overlaps.tsv"
        rev_file = f"{folder}/{program}_{target_tag}_to_{query_tag}_overlaps.tsv"

        if os.path.isfile(fwd_file) and os.path.isfile(rev_file):
            start = time.time()

            go_ahead = True

            fwd_df = pd.read_csv(fwd_file, sep="\t", encoding="utf-8", dtype=str, na_filter=False)

            duplicates = fwd_df[fwd_df.duplicated(subset=["gene_id_A", "gene_id_B"], keep=False)]

            if not duplicates.empty:
                print(f"Error: Duplicate query and target id pairs for fwd {program}:")
                print(duplicates)
                go_ahead = False

            rev_df = pd.read_csv(rev_file, sep="\t", encoding="utf-8", dtype=str, na_filter=False)

            duplicates = rev_df[rev_df.duplicated(subset=["gene_id_A", "gene_id_B"], keep=False)]

            if not duplicates.empty:
                print(f"Error: Duplicate query and target id pairs for rev {program}:")
                print(duplicates)
                go_ahead = False

            if go_ahead:

                fwd_df = fwd_df[fwd_df["gene_id_A"] != "NA"]
                rev_df = rev_df[rev_df["gene_id_A"] != "NA"]

                fwd_df = fwd_df[fwd_df["gene_id_B"] != "NA"]
                rev_df = rev_df[rev_df["gene_id_B"] != "NA"]

                fwd_df["gene_id_A_origin"] = fwd_df["gene_id_A_origin"].apply(clean_annotation_tag)
                rev_df["gene_id_B_origin"] = rev_df["gene_id_B_origin"].apply(clean_annotation_tag)

                fwd_df = fwd_df[fwd_df["gene_id_A_origin"] == self.name]
                rev_df = rev_df[rev_df["gene_id_B_origin"] == self.name]

                if liftoff:
                    equivalence_suffix = "liftoff_aegis"
                else:
                    equivalence_suffix = "lifton_aegis"

                fwd_hits = {}
            
                for index, row in fwd_df.iterrows():

                    gene_query = row["gene_id_A"]
                    gene_target = row["gene_id_B"]

                    copies = False
                    if row["gene_id_A_copy"] == "True":
                        copies = True
                        gene_query = gene_query.split("_")[:-1]
                        gene_query = "_".join(gene_query)

                    if row["gene_id_B_copy"] == "True":
                        copies = True
                        gene_target = gene_target.split("_")[:-1]
                        gene_target = "_".join(gene_target)

                    pair = f"{gene_query}---{gene_target}"

                    if synteny_present:
                        synteny = True
                        if row["gene_id_A_synteny_conserved"] == "False" or row["gene_id_B_synteny_conserved"] == "False":
                            synteny = False
                    else:
                        synteny = False

                    score = row["overlap_score"]

                    if synteny and copies:
                        score = f"{score} (synteny and copies)"
                    elif synteny:
                        score = f"{score} (synteny)"
                    elif copies:
                        score = f"{score} (copies)"

                    if pair not in fwd_hits:
                        fwd_hits[pair] = score
                    else:

                        old_score = int(fwd_hits[pair].split(" (")[0])
                        new_score = int(score.split(" (")[0])

                        if old_score > new_score:
                            new_score = old_score

                        if "multiple " in fwd_hits[pair]:
                            match_num = int(fwd_hits[pair].split("multiple ")[-1].split(")")[0])
                            match_num += 1
                        else:
                            match_num = 2

                        synteny = False
                        if "synteny" in fwd_hits[pair] or "synteny" in score:
                            synteny = True

                        copies = False
                        if "copies" in fwd_hits[pair] or "copies" in score:
                            copies = True
                        
                        if copies and synteny:
                            fwd_hits[pair] = f"{new_score} (synteny and copies) (multiple {match_num})"
                        elif copies:
                            fwd_hits[pair] = f"{new_score} (copies) (multiple {match_num})"
                        else:
                            raise ValueError(f"{pair} with score {score} should not have been already added to fwd_hits dictionary since at least one extra match should have been found through copies. The previous pair entry was already added with score {fwd_hits[pair]}. Query={query_tag} and Target={target_tag}.")

                rev_hits = {}
            
                for index, row in rev_df.iterrows():

                    gene_query = row["gene_id_B"]
                    gene_target = row["gene_id_A"]

                    copies = False
                    if row["gene_id_B_copy"] == "True":
                        copies = True
                        gene_query = gene_query.split("_")[:-1]
                        gene_query = "_".join(gene_query)

                    if row["gene_id_A_copy"] == "True":
                        copies = True
                        gene_target = gene_target.split("_")[:-1]
                        gene_target = "_".join(gene_target)

                    pair = f"{gene_query}---{gene_target}"

                    if synteny_present:
                        synteny = True
                        if row["gene_id_B_synteny_conserved"] == "False" or row["gene_id_A_synteny_conserved"] == "False":
                            synteny = False
                    else:
                        synteny = False

                    score = row["overlap_score"]

                    if synteny and copies:
                        score = f"{score} (synteny and copies)"
                    elif synteny:
                        score = f"{score} (synteny)"
                    elif copies:
                        score = f"{score} (copies)"

                    if pair not in rev_hits:
                        rev_hits[pair] = score
                    else:

                        old_score = int(rev_hits[pair].split(" (")[0])
                        new_score = int(score.split(" (")[0])

                        if old_score > new_score:
                            new_score = old_score

                        if "multiple " in rev_hits[pair]:
                            match_num = int(rev_hits[pair].split("multiple ")[-1].split(")")[0])
                            match_num += 1
                        else:
                            match_num = 2

                        synteny = False
                        if "synteny" in rev_hits[pair] or "synteny" in score:
                            synteny = True

                        copies = False
                        if "copies" in rev_hits[pair] or "copies" in score:
                            copies = True
                        
                        if copies and synteny:
                            rev_hits[pair] = f"{new_score} (synteny and copies) (multiple {match_num})"
                        elif copies:
                            rev_hits[pair] = f"{new_score} (copies) (multiple {match_num})"
                        else:
                            raise ValueError(f"{pair} with score {score} should not have been already added to rev_hits dictionary since at least one extra match should have been found through copies. The previous pair entry was already added with score {rev_hits[pair]}. Query={query_tag} and Target={target_tag}.")

                for fwd_pair in fwd_hits:

                    if fwd_pair in rev_hits:
                        equivalence_type = f"rec_{equivalence_suffix}"
                        score = f"{fwd_hits[fwd_pair]}/{rev_hits[fwd_pair]}"
                    else:
                        equivalence_type = f"fwd_{equivalence_suffix}"
                        score = f"{fwd_hits[fwd_pair]}"

                    gene_query = fwd_pair.split("---")[0]
                    gene_target = fwd_pair.split("---")[1]

                    self.genes[gene_query].equivalences.append(Equivalence(gene_target, equivalence_type, target_tag, species, score))


                for rev_pair in rev_hits:
                    if rev_pair not in fwd_hits:
                        equivalence_type = f"rev_{equivalence_suffix}"
                        score = f"{rev_hits[rev_pair]}"

                        gene_query = rev_pair.split("---")[0]
                        gene_target = rev_pair.split("---")[1]
                        
                        self.genes[gene_query].equivalences.append(Equivalence(gene_target, equivalence_type, target_tag, species, score))

                end = time.time()
                lapse = end - start
                if not quiet:
                    print(f"Adding {program} overlap equivalences for tags = [{query_tag}, {target_tag}] to {self.name} took {round(lapse/60, 2)} minutes")

        else:
            print(f"Warning: {fwd_file} or {rev_file} is missing.")


    def filter_equivalences(self, simple_rbh_blasts:bool=True, unidirectional_blasts:bool=True, replace:bool=True, identity_threshold=0, coverage_threshold=0, evalue_threshold=float("inf")):

        for gene in self.genes.values():
            gene.filter_equivalences(simple_rbh_blasts=simple_rbh_blasts, unidirectional_blasts=unidirectional_blasts, replace=replace, identity_threshold=identity_threshold, coverage_threshold=coverage_threshold, evalue_threshold=evalue_threshold)


def clean_annotation_tag(annotation_tag):

    annotation_tag = annotation_tag.replace("Lifton_", "")
    annotation_tag = annotation_tag.replace("Liftoff_", "")
    annotation_tag = annotation_tag.replace("Lifton_", "")
    annotation_tag = annotation_tag.replace("liftoff_", "")

    annotation_tag = annotation_tag.replace("Lifton", "")
    annotation_tag = annotation_tag.replace("Liftoff", "")
    annotation_tag = annotation_tag.replace("Lifton", "")
    annotation_tag = annotation_tag.replace("liftoff", "")

    annotation_tag = annotation_tag.split("_from_")[0]
    annotation_tag = annotation_tag.split("_on_")[0]
    annotation_tag = annotation_tag.split("_to_")[0]
    
    return annotation_tag